#	x509sak - The X.509 Swiss Army Knife white-hat certificate toolkit
#	Copyright (C) 2018-2020 Johannes Bauer
#
#	This file is part of x509sak.
#
#	x509sak is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	x509sak is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with x509sak; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import base64
from x509sak.OID import OID, OIDDB
from x509sak.estimate.BaseEstimator import BaseEstimator
from x509sak.estimate import JudgementCode, Commonness, Compatibility
from x509sak.estimate.Judgement import SecurityJudgement, SecurityJudgements, RFCReference
from x509sak.estimate.DistinguishedNameValidator import DistinguishedNameValidator

@BaseEstimator.register
class CertificateEstimator(BaseEstimator):
	_ALG_NAME = "certificate"
	_DN_VALIDATOR_SUBJECT = DistinguishedNameValidator.create_inherited("X509Cert_Body_Subject", validation_subject = "Subject")
	_DN_VALIDATOR_ISSUER = DistinguishedNameValidator.create_inherited("X509Cert_Body_Issuer", validation_subject = "Issuer")

	def _analyze_certificate_general_issues(self, certificate):
		judgements = SecurityJudgements()
		if certificate.version != 3:
			judgements += SecurityJudgement(JudgementCode.X509Cert_Body_Version_Not3, "Certificate version is v%d, usually would expect a v3 certificate." % (certificate.version), commonness = Commonness.HIGHLY_UNUSUAL)

		if certificate.serial < 0:
			standard = RFCReference(rfcno = 5280, sect = "4.1.2.2", verb = "MUST", text = "The serial number MUST be a positive integer assigned by the CA to each certificate.")
			judgements += SecurityJudgement(JudgementCode.X509Cert_Body_SerialNumber_BasicChecks_Negative, "Certificate serial number is a negative value.", compatibility = Compatibility.STANDARDS_DEVIATION, standard = standard)
		elif certificate.serial == 0:
			standard = RFCReference(rfcno = 5280, sect = "4.1.2.2", verb = "MUST", text = "The serial number MUST be a positive integer assigned by the CA to each certificate.")
			judgements += SecurityJudgement(JudgementCode.X509Cert_Body_SerialNumber_BasicChecks_Zero, "Certificate serial number is zero.", compatibility = Compatibility.STANDARDS_DEVIATION, standard = standard)
		elif certificate.serial >= (2 ** (8 * 20)):
			standard = RFCReference(rfcno = 5280, sect = "4.1.2.2", verb = "MUST", text = "Conforming CAs MUST NOT use serialNumber values longer than 20 octets.")
			judgements += SecurityJudgement(JudgementCode.X509Cert_Body_SerialNumber_BasicChecks_Large, "Certificate serial number is too large.", compatibility = Compatibility.STANDARDS_DEVIATION, standard = standard)

		if "non_der" in certificate.asn1_details.flags:
			judgements += SecurityJudgement(JudgementCode.X509Cert_Malformed_NonDEREncoding, "Certificate uses invalid DER encoding. Original certificate is %d bytes; DER would be %d bytes." % (len(certificate.asn1_details.original_der), len(certificate.asn1_details.encoded_der)), compatibility = Compatibility.STANDARDS_DEVIATION)

		if "trailing_data" in certificate.asn1_details.flags:
			judgements += SecurityJudgement(JudgementCode.X509Cert_TrailingData, "Certificate contains %d bytes of trailing data." % (len(certificate.asn1_details.tail)), compatibility = Compatibility.STANDARDS_DEVIATION)

		standard = RFCReference(rfcno = 5280, sect = [ "4.1.1.2", "4.1.2.3" ], verb = "MUST", text = "This field MUST contain the same algorithm identifier as the signature field in the sequence tbsCertificate (Section 4.1.2.3).")
		oid_header = OID.from_asn1(certificate.asn1["tbsCertificate"]["signature"]["algorithm"])
		oid_sig = OID.from_asn1(certificate.asn1["signatureAlgorithm"]["algorithm"])
		if oid_header != oid_sig:
			name_header = OIDDB.SignatureAlgorithms.get(oid_header, str(oid_header))
			name_sig = OIDDB.SignatureAlgorithms.get(oid_sig, str(oid_sig))
			judgements += SecurityJudgement(JudgementCode.X509Cert_Signature_Function_BodyMismatch, "Certificate indicates signature algorithm %s in header section and %s in signature section." % (name_header, name_sig), compatibility = Compatibility.STANDARDS_DEVIATION, standard = standard)
		else:
			# OIDs might be same, but parameters could differ (e.g., for RSA-PSS)
			header_hasvalue = certificate.asn1["tbsCertificate"]["signature"]["parameters"].hasValue()
			signature_hasvalue = certificate.asn1["signatureAlgorithm"]["parameters"].hasValue()
			if header_hasvalue and signature_hasvalue:
				name_header = OIDDB.SignatureAlgorithms.get(oid_header, str(oid_header))
				parameters_header = bytes(certificate.asn1["tbsCertificate"]["signature"]["parameters"])
				parameters_signature = bytes(certificate.asn1["signatureAlgorithm"]["parameters"])
				if parameters_header != parameters_signature:
					judgements += SecurityJudgement(JudgementCode.X509Cert_Signature_Function_BodyMismatch, "Certificate indicates same signature algorithm in both header section and signature section (%s), but parameterization of each differ." % (name_header), compatibility = Compatibility.STANDARDS_DEVIATION, standard = standard)
			elif header_hasvalue != signature_hasvalue:
				judgements += SecurityJudgement(JudgementCode.X509Cert_Signature_Function_BodyMismatch, "Certificate indicates same signature algorithm in both header section and signature section, but header %s while signature section %s." % ("has parameters" if header_hasvalue else "has no parameters", "does" if signature_hasvalue else "does not"), compatibility = Compatibility.STANDARDS_DEVIATION, standard = standard)

		return judgements

	def analyze(self, cert, root_cert = None):
		result = {
			"subject":		self.algorithm("dn").analyze(cert.subject, self._DN_VALIDATOR_SUBJECT),
			"issuer":		self.algorithm("dn").analyze(cert.issuer, self._DN_VALIDATOR_ISSUER),
			"validity":		self.algorithm("crt_validity").analyze(cert),
			"pubkey":		self.algorithm("pubkey").analyze(cert),
			"extensions":	self.algorithm("crt_exts").analyze(cert),
			"signature":	self.algorithm("sig").analyze(cert.signature_alg_oid, cert.signature_alg_params, cert.signature, root_cert),
			"purpose":		self.algorithm("purpose").analyze(cert),
			"security":		self._analyze_certificate_general_issues(cert),
		}
		if root_cert is not None:
			result["ca"] = self.algorithm("ca").analyze(cert, root_cert)
		if self._analysis_options.include_raw_data:
			result["raw"] = base64.b64encode(cert.der_data).decode("ascii")
		return result



#	x509sak - The X.509 Swiss Army Knife white-hat certificate toolkit
#	Copyright (C) 2018-2019 Johannes Bauer
#
#	This file is part of x509sak.
#
#	x509sak is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	x509sak is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with x509sak; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

from x509sak.estimate import AnalysisOptions

class BaseEstimator():
	_KNOWN_ALGORITHMS = { }
	_ALG_NAME = None

	def __init__(self, analysis_options = None):
		if analysis_options is None:
			analysis_options = AnalysisOptions()
		self._analysis_options = analysis_options

	@classmethod
	def register(cls, estimator_class):
		alg_name = estimator_class._ALG_NAME
		if alg_name is None:
			raise Exception("No algorithm name defined to register by.")
		if alg_name in cls._KNOWN_ALGORITHMS:
			raise Exception("Trying to re-register algorithm: %s" % (alg_name))
		cls._KNOWN_ALGORITHMS[alg_name] = estimator_class
		return estimator_class

	@classmethod
	def handler(cls, alg_name):
		if alg_name not in cls._KNOWN_ALGORITHMS:
			raise KeyError("Algorithm quality of '%s' cannot be estimated, Estimator class not registered." % (alg_name))
		return cls._KNOWN_ALGORITHMS[alg_name]

	def algorithm(self, alg_name):
		handler = self.handler(alg_name)
		return handler(analysis_options = self._analysis_options)

	def analyze(self, *args, **kwargs):
		raise NotImplementedError("method 'analyze'")



#	x509sak - The X.509 Swiss Army Knife white-hat certificate toolkit
#	Copyright (C) 2018-2019 Johannes Bauer
#
#	This file is part of x509sak.
#
#	x509sak is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	x509sak is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with x509sak; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import os
import random
import logging
from x509sak.SubprocessExecutor import SubprocessExecutor

_log = logging.getLogger("x509sak.PrimeDB")

class PrimeDB():
	def __init__(self, directory = ".", generator_program = None):
		self._directory = directory
		self._generator = generator_program
		if not self._directory.endswith("/"):
			self._directory += "/"
		self._primes = { }

	def _get_filename(self, bitlen, primetype):
		assert(primetype in [ "2msb", "3msb" ])
		filename = self._directory + "primes_%s_%d.txt" % (primetype, bitlen)
		return filename

	def _load_primes(self, bitlen, primetype, offset = 0):
		filename = self._get_filename(bitlen, primetype)
		primes = [ ]
		with open(filename) as f:
			f.seek(offset)
			for line in f:
				if line.startswith("#"):
					continue
				p = int(line.rstrip("\r\n"), 16)
				primes.append(p)
			random.shuffle(primes)
		self._primes[(bitlen, primetype)] = primes

	def add_prime(self, prime):
		if prime.bit_length() < 8:
			raise NotImplementedError("We're not interesting in tiny primes.")
		msb_2_set = ((prime >> (prime.bit_length() - 2)) & 1) == 1
		msb_3_set = ((prime >> (prime.bit_length() - 3)) & 1) == 1
		if msb_2_set and msb_3_set:
			prime_type = "3msb"
		elif msb_2_set:
			prime_type = "2msb"
		else:
			prime_type = "1msb"
		key = (prime.bit_length(), prime_type)
		if key not in self._primes:
			self._primes[key] = [ ]
		self._primes[key].append(prime)
		return self

	def add(self, *primes):
		for prime in primes:
			self.add_prime(prime)
		return self

	def write(self, overwrite_files = False):
		for ((bitlen, primetype), primes) in self._primes.items():
			filename = self._get_filename(bitlen, primetype)
			if (len(primes) > 0) and (overwrite_files or (not os.path.exists(filename))):
				with open(filename, "w") as f:
					for prime in primes:
						print("%x" % (prime), file = f)

	def _generate_primes(self, bitlen, primetype):
		filename = self._get_filename(bitlen, primetype)
		if os.path.isfile(filename):
			offset = os.stat(filename).st_size
		else:
			offset = 0
		if bitlen <= 1024:
			count = 100
		elif bitlen <= 2048:
			count = 10
		elif bitlen <= 8192:
			count = 5
		else:
			count = 1
		_log.debug("Generating %d %d bit %s prime(s) using %s; current DB %s offset is %d", count, bitlen, primetype, self._generator, filename, offset)
		cmd = [ self._generator, "--prime-type", primetype, "--bit-length", str(bitlen), "--num-primes", str(count) ]
		SubprocessExecutor(cmd).run()

		# finally re-read database from previous offset
		self._load_primes(bitlen, primetype, offset)

	def get(self, bitlen, primetype = "2msb"):
		key = (bitlen, primetype)
		if key not in self._primes:
			try:
				self._load_primes(bitlen, primetype)
			except FileNotFoundError:
				pass
		if (key not in self._primes) or (len(self._primes[key]) == 0):
			# Prime DB is exhausted.
			if self._generator is not None:
				# Generate new primes on-the-fly and try again
				self._generate_primes(bitlen, primetype)
				return self.get(bitlen = bitlen, primetype = primetype)
			else:
				raise Exception("Primes of type %s with bitlength %d exhausted in database. Generate these primes and save them in a file called \"%s\". Execute: primegen/primegen -b %d -p %s" % (primetype, bitlen, self._get_filename(bitlen, primetype), bitlen, primetype))
		return self._primes[key].pop()



#	x509sak - The X.509 Swiss Army Knife white-hat certificate toolkit
#	Copyright (C) 2018-2018 Johannes Bauer
#
#	This file is part of x509sak.
#
#	x509sak is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	x509sak is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with x509sak; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

from x509sak.tests.BaseTest import BaseTest
from x509sak.KwargsChecker import KwargsChecker
from x509sak.Exceptions import InvalidInputException

class KwargsCheckerTests(BaseTest):
	def test_normal(self):
		checker = KwargsChecker(required_arguments = set([ "foo", "bar" ]))
		checker.check({ "foo": 2, "bar": 9 })

		checker = KwargsChecker(required_arguments = set([ "foo", "bar" ]), optional_arguments = set([ "moo" ]))
		checker.check({ "foo": 2, "bar": 9 })
		checker.check({ "foo": 2, "bar": 9, "moo": 12 })

		checker = KwargsChecker(optional_arguments = set([ "moo" ]))
		checker.check({ })
		checker.check({ "moo": 9 })

	def test_missing_required(self):
		checker = KwargsChecker(required_arguments = set([ "foo", "bar" ]))
		with self.assertRaises(InvalidInputException):
			checker.check({ "foo": 2 })
		with self.assertRaises(InvalidInputException):
			checker.check({ "foo": 2, "BAR": 9 })
		with self.assertRaises(InvalidInputException):
			checker.check({ "foo": 2, "BAR": 9, "MOO": 12 })
		with self.assertRaises(InvalidInputException):
			checker.check({ "foo": 2, "BAR": 9, "MOO": 12 }, hint = "foobar")

		checker = KwargsChecker(required_arguments = set([ "foo", "bar" ]), optional_arguments = set([ "muh" ]))
		with self.assertRaises(InvalidInputException):
			checker.check({ "foo": 2, "BAR": 9, "MOO": 12 }, hint = "foobar")

	def test_single(self):
		checker = KwargsChecker(required_arguments = set([ "foo", "bar" ]), optional_arguments = set([ "moo" ]))
		checker.check_single("foo")
		checker.check_single("bar")
		checker.check_single("moo")
		with self.assertRaises(InvalidInputException):
			checker.check_single("fux")
		with self.assertRaises(InvalidInputException):
			checker.check_single("fux", hint = "foobarhint")



#	x509sak - The X.509 Swiss Army Knife white-hat certificate toolkit
#	Copyright (C) 2018-2018 Johannes Bauer
#
#	This file is part of x509sak.
#
#	x509sak is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	x509sak is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with x509sak; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import collections
from x509sak.tests import BaseTest

class MultiplicationBitsTests(BaseTest):
	@staticmethod
	def _enumerate_bit_values(bits):
		return range(2 ** (bits - 1), 2 ** bits)

	@staticmethod
	def _set_msb_bits(value, bitcnt, bitvalue):
		assert(bitcnt <= bitvalue.bit_length())
		bits = value.bit_length()
		shift = max((bits - bitcnt), 0)
		mask = ((1 << bitcnt) - 1) << shift
		value = value & ~mask
		value = value | (bitvalue << shift)
		return value

	def _set_msb_bits_11(self, value):
		return self._set_msb_bits(value, 2, 0b11)

	def _set_msb_bits_111(self, value):
		return self._set_msb_bits(value, 3, 0b111)

	def test_msb_bits(self):
		self.assertEqual(self._set_msb_bits(0b100000000, 3, 0b101), 0b101000000)
		self.assertEqual(self._set_msb_bits(0b100000000, 4, 0b1011), 0b101100000)
		self.assertEqual(self._set_msb_bits(0b100000000, 6, 0b101101), 0b101101000)
		self.assertEqual(self._set_msb_bits(0, 3, 0b101), 0b101)

	def test_enumerate_bitvalues(self):
		bit = 4
		values = list(self._enumerate_bit_values(bit))
		bitlengths = list(set(value.bit_length() for value in values))
		self.assertEqual(bitlengths, [ bit ])
		self.assertEqual((values[0] - 1).bit_length(), bit - 1)
		self.assertEqual(values[0].bit_length(), bit)
		self.assertEqual(values[-1].bit_length(), bit)
		self.assertEqual((values[-1] + 1).bit_length(), bit + 1)

	def test_multiply_two(self):
		bit = 4
		values = self._enumerate_bit_values(4)
		for p in values:
			for q in values:
				if q < p:
					continue
				n = (self._set_msb_bits_11(p)) * (self._set_msb_bits_11(q))
				self.assertEqual(n.bit_length(), 2 * bit)

	def test_multiply_any_three(self):
		bit = 4
		values = self._enumerate_bit_values(bit)

		bits = collections.Counter()
		for p in values:
			for q in values:
				for r in values:
					q_ = (2 * self._set_msb_bits_111(p) * (self._set_msb_bits_111(q))) + 1
					n = q_ * self._set_msb_bits_111(r)
					length = n.bit_length()
					bits[length] += 1
		self.assertEqual(len(bits), 1)

	def test_multiply_pq_three(self):
		bit = 4
		values = self._enumerate_bit_values(bit)

		bits = collections.Counter()
		for p in values:
			for q in values:
				p = self._set_msb_bits(p, 3, 0b111)
				q = self._set_msb_bits(q, 3, 0b111)
				q_ = (2 * p * q) + 1
				n = p * q_
				length = n.bit_length()
				bits[length] += 1
		self.assertEqual(len(bits), 1)



#	x509sak - The X.509 Swiss Army Knife white-hat certificate toolkit
#	Copyright (C) 2018-2020 Johannes Bauer
#
#	This file is part of x509sak.
#
#	x509sak is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	x509sak is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with x509sak; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import urllib.parse
from x509sak.estimate.Judgement import Commonness, Compatibility
from x509sak.ASN1Wrapper import ASN1GeneralNameWrapper
from x509sak.Tools import ValidationTools
from x509sak.estimate.Validator import BaseValidator, BaseValidationResult
from x509sak.IPAddress import IPAddressSubnet

class GeneralNameValidationResult(BaseValidationResult):
	_PRIVATE_SUBNETS_IPV4 = (
		("current network",			IPAddressSubnet.from_str("0.0.0.0/8")),
		("private class A",			IPAddressSubnet.from_str("10.0.0.0/8")),
		("private class B",			IPAddressSubnet.from_str("172.16.0.0/12")),
		("private class C",			IPAddressSubnet.from_str("192.168.0.0/16")),
		("carrier-grade NAT",		IPAddressSubnet.from_str("100.64.0.0/10")),
		("loopback",				IPAddressSubnet.from_str("127.0.0.0/8")),
		("link-local address",		IPAddressSubnet.from_str("169.254.0.0/16")),
		("benchmarking subnet",		IPAddressSubnet.from_str("198.18.0.0/15")),
		("IP multicast",			IPAddressSubnet.from_str("224.0.0.0/4")),
		("reserved",				IPAddressSubnet.from_str("240.0.0.0/4")),
		("limited broadcast",		IPAddressSubnet.from_str("255.255.255.255/32")),
	)

	_PRIVATE_SUBNETS_IPV6 = (
		("loopback",				IPAddressSubnet.from_str("::1/128")),
		("discard",					IPAddressSubnet.from_str("100::/64")),
		("deprecated 6to4 scheme",	IPAddressSubnet.from_str("2002::/16")),
		("link-local address",		IPAddressSubnet.from_str("fc00::/7")),
		("link-local address",		IPAddressSubnet.from_str("fe80::/8")),
		("multicast address",		IPAddressSubnet.from_str("ff00::/8")),
	)

	def _get_message(self, issue, message):
		return "%s of type %s %s" % (self._validator.validation_subject, self._subject.name, message)

	def _validate_dNSName(self):
		self._report("Enc_DER_Struct_GenName_DNS_Unexpected", "contains unexpected domain name \"%s\"." % (self._subject.str_value), commonness = Commonness.UNUSUAL)
		if self._subject.str_value == " ":
			return self._report("Enc_DER_Struct_GenName_DNS_OnlyWhitespace", "got invalid DNS name \" \" (space character).", commonness = Commonness.HIGHLY_UNUSUAL, compatibility = Compatibility.STANDARDS_DEVIATION)

		if self._validator.allow_dnsname_wildcard_matches:
			(result, label) = ValidationTools.validate_domainname_template(self._subject.str_value)
			if result != ValidationTools.DomainnameTemplateValidationResult.Valid:
				if result == ValidationTools.DomainnameTemplateValidationResult.InvalidCharacter:
					return self._report("Enc_DER_Struct_GenName_DNS_Malformed", "has invalid domain name \"%s\", error at label \"%s\"." % (self._subject.str_value, label), commonness = Commonness.HIGHLY_UNUSUAL, compatibility = Compatibility.STANDARDS_DEVIATION)
				elif result == ValidationTools.DomainnameTemplateValidationResult.FullWildcardNotLeftmost:
					return self._report("Enc_DER_Struct_GenName_DNS_Wildcard_NotLeftmost", "has invalid domain name \"%s\". Full-label wildcard appears not as leftmost element." % (self._subject.str_value), commonness = Commonness.UNUSUAL, compatibility = Compatibility.STANDARDS_DEVIATION)
				elif result == ValidationTools.DomainnameTemplateValidationResult.MoreThanOneWildcard:
					return self._report("Enc_DER_Struct_GenName_DNS_Wildcard_MulitpleWildcards", "has invalid domain name \"%s\". More than one wildcard label present." % (self._subject.str_value), commonness = Commonness.UNUSUAL, compatibility = Compatibility.STANDARDS_DEVIATION)
				elif result == ValidationTools.DomainnameTemplateValidationResult.WildcardInInternationalDomain:
					return self._report("Enc_DER_Struct_GenName_DNS_Wildcard_InternationalLabel", "has invalid domain name \"%s\". Wildcard in international domain label \"%s\"." % (self._subject.str_value, label), commonness = Commonness.HIGHLY_UNUSUAL, compatibility = Compatibility.STANDARDS_DEVIATION)
				else:
					raise NotImplementedError(result)

			if "*" in self._subject.str_value:
				# Wildcard match
				labels = self._subject.str_value.split(".")
				if len(labels) <= 2:
					self._report("Enc_DER_Struct_GenName_DNS_Wildcard_BroadMatch", "has wildcard value \"%s\", which is an extremely broad domain match." % (self._subject.str_value), commonness = Commonness.HIGHLY_UNUSUAL)

		else:
			if "*" in self._subject.str_value:
				# Not permitted but wildcard present
				self._report("Enc_DER_Struct_GenName_DNS_Wildcard_NotPermitted", "has wildcard value \"%s\", which is not permitted for this type." % (self._subject.str_value), commonness = Commonness.HIGHLY_UNUSUAL)

		if not "." in self._subject.str_value:
			self._report("Enc_DER_Struct_GenName_DNS_SingleLabel", "contains only single label \"%s\", which is highly unusual." % (self._subject.str_value), commonness = Commonness.HIGHLY_UNUSUAL)

		validation_name = self._subject.str_value
		if self._validator.allow_dnsname_wildcard_matches:
			validation_name = validation_name.replace("*", "a")
		result = ValidationTools.validate_domainname(validation_name)
		if not result:
			self._report("Enc_DER_Struct_GenName_DNS_Malformed", "has invalid domain name \"%s\" (wildcard matches %s)." % (self._subject.str_value, "permitted" if self._validator.allow_dnsname_wildcard_matches else "forbidden"), commonness = Commonness.HIGHLY_UNUSUAL, compatibility = Compatibility.STANDARDS_DEVIATION)

	def _validate_iPAddress(self):
		self._report("Enc_DER_Struct_GenName_IPAddress_Unexpected", "contains unexpected IP address \"%s\"." % (self._subject.str_value), commonness = Commonness.UNUSUAL)

		address_length = len(self._subject.asn1_value)
		if (not self._validator.ip_addresses_are_subnets) and (address_length not in [ 4, 16 ]):
			self._report("Enc_DER_Struct_GenName_IPAddress_Malformed", "expects either 4 or 16 bytes of data for IPv4/IPv6, but saw %d bytes." % (len(self._subject.str_value)), commonness = Commonness.HIGHLY_UNUSUAL, compatibility = Compatibility.STANDARDS_DEVIATION)
		elif self._validator.ip_addresses_are_subnets and (address_length not in [ 8, 32 ]):
			self._report("Enc_DER_Struct_GenName_IPAddress_Malformed", "expects either 8 or 32 bytes of data for IPv4/IPv6 subnet, but saw %d bytes." % (len(self._subject.str_value)), commonness = Commonness.HIGHLY_UNUSUAL, compatibility = Compatibility.STANDARDS_DEVIATION)
		else:
			subject = self._subject.ip
			ipv4 = subject.is_ipv4
			private_subnets = self._PRIVATE_SUBNETS_IPV4 if ipv4 else self._PRIVATE_SUBNETS_IPV6
			for (network_class, subnet) in private_subnets:
				if subnet.overlaps(subject):
					self._report("Enc_DER_Struct_GenName_IPAddress_PrivateAddressSpace", "has network address %s in a %s subnet." % (str(subject), network_class), commonness = Commonness.UNUSUAL)
					break

		if self._validator.ip_addresses_are_subnets and (address_length in [ 8, 32 ]):
			if self._subject.ip.overlap:
				self._report("Enc_DER_Struct_GenName_IPAddress_MalformedSubnet", "has network address bits set which are masked by subnet: %s" % (self._subject.ip), commonness = Commonness.HIGHLY_UNUSUAL)

			if not self._subject.ip.is_cidr:
				self._report("Enc_DER_Struct_GenName_IPAddress_NonCIDRBlock", "has subnet %s that cannot be expressed as CIDR block" % (self._subject.ip), commonness = Commonness.UNUSUAL)

	def _validate_rfc822Name(self):
		self._report("Enc_DER_Struct_GenName_Email_Unexpected", "contains unexpected email address \"%s\"." % (self._subject.str_value), commonness = Commonness.UNUSUAL)
		if not ValidationTools.validate_email_address(self._subject.str_value):
			self._report("Enc_DER_Struct_GenName_Email_Malformed", "contains invalid email address \"%s\"." % (self._subject.str_value), commonness = Commonness.HIGHLY_UNUSUAL, compatibility = Compatibility.STANDARDS_DEVIATION)

	def _validate_directoryName(self):
		self._report("Enc_DER_Struct_GenName_DirectoryAddress_Unexpected", "contains unexpected directory name \"%s\"." % (self._subject.str_value), commonness = Commonness.UNUSUAL)
		if self._subject.directory_name.rdn_count == 0:
			self._report("Enc_DER_Struct_GenName_DirectoryAddress_Empty", "contains empty directory name \"%s\"." % (self._subject.str_value), commonness = Commonness.HIGHLY_UNUSUAL)

	def _validate_uniformResourceIdentifier(self):
		self._report("Enc_DER_Struct_GenName_URI_Unexpected", "contains unexpected URI \"%s\"." % (self._subject.str_value))
		if not ValidationTools.validate_uri(str(self._subject.str_value)):
			self._report("Enc_DER_Struct_GenName_URI_Malformed", "contains invalid URI \"%s\"." % (str(self._subject.str_value)), commonness = Commonness.HIGHLY_UNUSUAL, compatibility = Compatibility.STANDARDS_DEVIATION)
		if self._validator.permissible_uri_schemes is not None:
			split_url = urllib.parse.urlsplit(self._subject.str_value)
			if split_url.scheme not in self._validator.permissible_uri_schemes:
				self._report("Enc_DER_Struct_GenName_URI_UncommonURIScheme", "contains invalid URI scheme \"%s\" (permitted schemes are only %s)." % (str(self._subject.str_value), ", ".join(sorted(self._validator.permissible_uri_schemes))), commonness = Commonness.UNUSUAL)

	def _validate_registeredID(self):
		self._report("Enc_DER_Struct_GenName_RegisteredID_Unexpected", "contains unexpected registered ID \"%s\"." % (self._subject.str_value), commonness = Commonness.UNUSUAL)

	def _validate_otherName(self):
		self._report("Enc_DER_Struct_GenName_OtherName_Unexpected", "contains unexpected other name \"%s\"." % (self._subject.str_value), commonness = Commonness.HIGHLY_UNUSUAL)

	def _validate_x400Address(self):
		self._report("Enc_DER_Struct_GenName_X400Address_Unexpected", "contains unexpected X.400 address \"%s\"." % (self._subject.str_value), commonness = Commonness.HIGHLY_UNUSUAL)

	def _validate_ediPartyName(self):
		self._report("Enc_DER_Struct_GenName_EDIPartyName_Unexpected", "contains unexpected EDI party name \"%s\"." % (self._subject.str_value), commonness = Commonness.HIGHLY_UNUSUAL)

	def _validate(self):
		method_name = "_validate_%s" % (str(self._subject.name))
		gn_subtype_handler = getattr(self, method_name, None)
		if gn_subtype_handler is not None:
			gn_subtype_handler()
		else:
			print(method_name)

class GeneralNameValidator(BaseValidator):
	_ValidationResultClass = GeneralNameValidationResult

	def __init__(self, validation_subject, recognized_issues, permissible_uri_schemes = None, allow_dnsname_wildcard_matches = False, ip_addresses_are_subnets = False):
		BaseValidator.__init__(self, validation_subject, recognized_issues)
		self._permissible_uri_schemes = permissible_uri_schemes
		self._allow_dnsname_wildcard_matches = allow_dnsname_wildcard_matches
		self._ip_addresses_are_subnets = ip_addresses_are_subnets

	@property
	def permissible_uri_schemes(self):
		return self._permissible_uri_schemes

	@property
	def allow_dnsname_wildcard_matches(self):
		return self._allow_dnsname_wildcard_matches

	@property
	def ip_addresses_are_subnets(self):
		return self._ip_addresses_are_subnets

	def validate_asn1(self, general_name_asn1):
		general_name = ASN1GeneralNameWrapper.from_asn1(general_name_asn1)
		return self.validate(general_name)



#	x509sak - The X.509 Swiss Army Knife white-hat certificate toolkit
#	Copyright (C) 2020-2020 Johannes Bauer
#
#	This file is part of x509sak.
#
#	x509sak is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	x509sak is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with x509sak; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

from x509sak.OID import OIDDB
from x509sak.estimate.BaseEstimator import BaseEstimator
from x509sak.estimate import JudgementCode, Compatibility
from x509sak.estimate.Judgement import SecurityJudgement, SecurityJudgements, Commonness, Verdict, RFCReference
from x509sak.estimate.DistinguishedNameValidator import DistinguishedNameValidator
from x509sak.Tools import TextTools

@BaseEstimator.register
class CARelationshipSecurityEstimator(BaseEstimator):
	_ALG_NAME = "ca"
	_DN_VALIDATOR_CA_SUBJECT = DistinguishedNameValidator.create_inherited("X509Cert_Body_Subject", validation_subject = "CA subject")
	_DN_VALIDATOR_CA_ISSUER = DistinguishedNameValidator.create_inherited("X509Cert_Body_Issuer", validation_subject = "CA issuer")

	def _judge_validity_timestamps(self, certificate, ca_certificate):
		judgements = SecurityJudgements()
		if any(ts is None for ts in [ certificate.valid_not_before, certificate.valid_not_after, ca_certificate.valid_not_before, ca_certificate.valid_not_after ]):
			judgements += SecurityJudgement(JudgementCode.CertUsage_CARelationship_Validity_MalformedTimestamp, "Certificate or CA certificate validity field malformed, cannot perform checking of time intervals.", verdict = Verdict.NO_SECURITY, commonness = Commonness.HIGHLY_UNUSUAL)
		else:
			if certificate.valid_not_before > ca_certificate.valid_not_after:
				judgements += SecurityJudgement(JudgementCode.CertUsage_CARelationship_Validity_NoOverlap, "Certifiate becomes valid after CA certificate has expired.", verdict = Verdict.NO_SECURITY, commonness = Commonness.HIGHLY_UNUSUAL)
			elif certificate.valid_not_after < ca_certificate.valid_not_before:
				judgements += SecurityJudgement(JudgementCode.CertUsage_CARelationship_Validity_NoOverlap, "Certifiate becomes invalid before CA certificate becomes valid.", verdict = Verdict.NO_SECURITY, commonness = Commonness.HIGHLY_UNUSUAL)
			elif ca_certificate.valid_not_before <= certificate.valid_not_before <= certificate.valid_not_after <= ca_certificate.valid_not_after:
				judgements += SecurityJudgement(JudgementCode.CertUsage_CARelationship_Validity_FullOverlap, "Certifiate validity interval falls fully into CA certificate validity.", commonness = Commonness.COMMON)
			else:
				judgements += SecurityJudgement(JudgementCode.CertUsage_CARelationship_Validity_PartialOverlap, "Certifiate validity interval falls partially into CA certificate validity. The CA certificate therefore is only useful for a portion of lifetime of the certificate under test.", commonness = Commonness.UNUSUAL)

		return judgements

	def _judge_signature(self, certificate, ca_certificate):
		judgements = SecurityJudgements()

		if not ca_certificate.is_ca_certificate:
			judgements += SecurityJudgement(JudgementCode.CertUsage_CARelationship_CACertificateInvalidAsCA, "CA certificate is not valid as a CA.", verdict = Verdict.NO_SECURITY, commonness = Commonness.HIGHLY_UNUSUAL)

		if not certificate.signed_by(ca_certificate):
			judgements += SecurityJudgement(JudgementCode.CertUsage_CARelationship_Signature_VerificationFailure, "Certificate signature is not verifiable with CA public key.", verdict = Verdict.NO_SECURITY, commonness = Commonness.HIGHLY_UNUSUAL)
		else:
			judgements += SecurityJudgement(JudgementCode.CertUsage_CARelationship_Signature_VerificationSuccess, "Certificate signature is valid.", commonness = Commonness.COMMON)

		return judgements

	def analyze(self, certificate, ca_certificate):
		judgements = SecurityJudgements()

		if certificate.issuer != ca_certificate.subject:
			standard = RFCReference(rfcno = 5280, sect = "4.1.2.4", verb = "MUST", text = "The issuer field identifies the entity that has signed and issued the certificate.")
			judgements += SecurityJudgement(JudgementCode.CertUsage_CARelationship_Subject_Issuer_Mismatch, "Certificate issuer does not match CA subject.", commonness = Commonness.HIGHLY_UNUSUAL, compatibility = Compatibility.STANDARDS_DEVIATION, standard = standard)
		else:
			judgements += SecurityJudgement(JudgementCode.CertUsage_CARelationship_Subject_Issuer_Match, "Certificate issuer matches CA subject.", commonness = Commonness.COMMON)

		aki = certificate.extensions.get_first(OIDDB.X509Extensions.inverse("AuthorityKeyIdentifier"))
		ski = ca_certificate.extensions.get_first(OIDDB.X509Extensions.inverse("SubjectKeyIdentifier"))

		if aki is not None:
			if aki.keyid is not None:
				if ski is None:
					standard = RFCReference(rfcno = 5280, sect = "4.2.1.2", verb = "MUST", text = "In conforming CA certificates, the value of the subject key identifier MUST be the value placed in the key identifier field of the authority key identifier extension (Section 4.2.1.1) of certificates issued by the subject of this certificate.")
					judgements += SecurityJudgement(JudgementCode.CertUsage_CARelationship_AKI_KeyID_Uncheckable, "Key ID specified in the Authority Key Identifier extension of the certificate cannot be validated because CA does not have a Subject Key Identifier extension.", commonness = Commonness.HIGHLY_UNUSUAL)
				else:
					if ski.keyid != aki.keyid:
						standard = RFCReference(rfcno = 5280, sect = "4.2.1.2", verb = "MUST", text = "In conforming CA certificates, the value of the subject key identifier MUST be the value placed in the key identifier field of the authority key identifier extension (Section 4.2.1.1) of certificates issued by the subject of this certificate.")
						judgements += SecurityJudgement(JudgementCode.CertUsage_CARelationship_AKI_KeyID_Mismatch, "Key ID specified in the Authority Key Identifier extension of the certificate does not match the Key ID of the CA Subject Key Identifier extension.", commonness = Commonness.HIGHLY_UNUSUAL)
					else:
						judgements += SecurityJudgement(JudgementCode.CertUsage_CARelationship_AKI_KeyID_Match, "Key ID specified in the Authority Key Identifier extension matches that of the CA Subject Key Identifier extension.", commonness = Commonness.COMMON)

			if aki.ca_names is not None:
				dir_name_count = 0
				for ca_name in aki.ca_names:
					if ca_name.name == "directoryName":
						dir_name_count += 1
						if ca_name.directory_name == ca_certificate.subject:
							judgements += SecurityJudgement(JudgementCode.CertUsage_CARelationship_AKI_CAName_Match, "Directory name specified in the Authority Key Identifier extension of the certificate matches the CA subject.", commonness = Commonness.COMMON)
							break
				else:
					names = "None of the %d CA names" % (len(aki.ca_names)) if (len(aki.ca_names) != 1) else "The CA name"
					judgements += SecurityJudgement(JudgementCode.CertUsage_CARelationship_AKI_CAName_Mismatch, "%s (%s) specified in the Authority Key Identifier extension of the certificate matches the CA subject." % (names, TextTools.sp(dir_name_count, "directoryName")), commonness = Commonness.HIGHLY_UNUSUAL, compatibility = Compatibility.STANDARDS_DEVIATION)

			if aki.serial is not None:
				if aki.serial != ca_certificate.serial:
					judgements += SecurityJudgement(JudgementCode.CertUsage_CARelationship_AKI_Serial_Mismatch, "Serial number specified in the Authority Key Identifier extension of the certificate does not match the serial number of the CA certificate.", commonness = Commonness.HIGHLY_UNUSUAL)
				else:
					judgements += SecurityJudgement(JudgementCode.CertUsage_CARelationship_AKI_Serial_Match, "Serial number specified in the Authority Key Identifier extension matches that of the CA certificate.", commonness = Commonness.COMMON)

		judgements += self._judge_validity_timestamps(certificate, ca_certificate)
		judgements += self._judge_signature(certificate, ca_certificate)

		return {
			"ca_subject":	self.algorithm("dn").analyze(ca_certificate.subject, self._DN_VALIDATOR_CA_SUBJECT),
			"ca_issuer":	self.algorithm("dn").analyze(ca_certificate.issuer, self._DN_VALIDATOR_CA_ISSUER),
			"security":		judgements,
		}



#	x509sak - The X.509 Swiss Army Knife white-hat certificate toolkit
#	Copyright (C) 2018-2020 Johannes Bauer
#
#	This file is part of x509sak.
#
#	x509sak is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	x509sak is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with x509sak; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

from pyasn1.type.univ import BitString, Integer, OctetString
from pyasn1.type.useful import UTCTime
from pyasn1.type.constraint import ValueRangeConstraint
from x509sak.tests.BaseTest import BaseTest
from x509sak.Tools import ASN1Tools

class BasicASN1Model(Integer): pass
class RestrictiveASN1Model(Integer):
	subtypeSpec = ValueRangeConstraint(100, 199)

class ASN1ToolTests(BaseTest):
	def test_bitstring_trailing_zero(self):
		self.assertTrue(ASN1Tools.bitstring_has_trailing_zeros(BitString([ 0, 1, 0 ])))
		self.assertTrue(ASN1Tools.bitstring_has_trailing_zeros(BitString([ 0 ])))
		self.assertTrue(ASN1Tools.bitstring_has_trailing_zeros(BitString([ 1, 1, 1, 1, 1, 1, 1, 0 ])))
		self.assertTrue(ASN1Tools.bitstring_has_trailing_zeros(BitString([ 1, 1, 1, 1, 1, 1, 1, 1, 0 ])))

		self.assertFalse(ASN1Tools.bitstring_has_trailing_zeros(BitString([ 0, 1 ])))
		self.assertFalse(ASN1Tools.bitstring_has_trailing_zeros(BitString([ ])))
		self.assertFalse(ASN1Tools.bitstring_has_trailing_zeros(BitString([ 0, 0, 0, 0, 1 ])))
		self.assertFalse(ASN1Tools.bitstring_has_trailing_zeros(BitString([ 0, 0, 0, 0, 0, 0, 1 ])))
		self.assertFalse(ASN1Tools.bitstring_has_trailing_zeros(BitString([ 0, 0, 0, 0, 0, 0, 0, 1 ])))

	def test_bitstring_empty(self):
		self.assertTrue(ASN1Tools.bitstring_is_empty(BitString([ ])))
		self.assertTrue(ASN1Tools.bitstring_is_empty(BitString([ 0 ])))
		self.assertTrue(ASN1Tools.bitstring_is_empty(BitString([ 0, 0, 0 ])))

		self.assertFalse(ASN1Tools.bitstring_is_empty(BitString([ 1 ])))
		self.assertFalse(ASN1Tools.bitstring_is_empty(BitString([ 0, 1 ])))
		self.assertFalse(ASN1Tools.bitstring_is_empty(BitString([ 0, 0, 1, 0 ])))

	def test_bitstring_highest_set_bit(self):
		self.assertEqual(ASN1Tools.bitstring_highbit(BitString([ ])), None)
		self.assertEqual(ASN1Tools.bitstring_highbit(BitString([ 0 ])), None)
		self.assertEqual(ASN1Tools.bitstring_highbit(BitString([ 0, 0 ])), None)
		self.assertEqual(ASN1Tools.bitstring_highbit(BitString([ 1 ])), 0)
		self.assertEqual(ASN1Tools.bitstring_highbit(BitString([ 1, 1 ])), 1)
		self.assertEqual(ASN1Tools.bitstring_highbit(BitString([ 1, 1, 1 ])), 2)
		self.assertEqual(ASN1Tools.bitstring_highbit(BitString([ 1, 1, 1, 0 ])), 2)
		self.assertEqual(ASN1Tools.bitstring_highbit(BitString([ 1, 1, 1, 0, 0 ])), 2)
		self.assertEqual(ASN1Tools.bitstring_highbit(BitString([ 1, 1, 1, 0, 0, 1 ])), 5)
		self.assertEqual(ASN1Tools.bitstring_highbit(BitString([ 1, 1, 1, 0, 0, 1, 0 ])), 5)
		self.assertEqual(ASN1Tools.bitstring_highbit(BitString([ 1, 1, 1, 0, 0, 1, 0, 1 ])), 7)

	def test_safedecode_undecodable(self):
		der_data = bytes.fromhex("aa bb cc")
		asn1_details = ASN1Tools.safe_decode(der_data)
		self.assertEqual(asn1_details.flags, set([ "undecodable" ]))
		self.assertEqual(asn1_details.tail, None)
		self.assertEqual(asn1_details.asn1, None)
		self.assertEqual(asn1_details.generic_asn1, None)
		self.assertEqual(asn1_details.encoded_der, None)
		self.assertEqual(asn1_details.original_der, der_data)
		self.assertEqual(asn1_details.model_index, None)

	def test_safedecode_ok_nomodel(self):
		der_data = bytes.fromhex("02 01   7b")		# 123
		asn1_details = ASN1Tools.safe_decode(der_data)
		self.assertEqual(asn1_details.flags, set())
		self.assertEqual(asn1_details.tail, bytes())
		self.assertEqual(asn1_details.asn1, Integer(123))
		self.assertEqual(asn1_details.generic_asn1, None)
		self.assertEqual(asn1_details.encoded_der, None)
		self.assertEqual(asn1_details.original_der, der_data)
		self.assertEqual(asn1_details.model_index, 0)

	def test_safedecode_ok_model(self):
		der_data = bytes.fromhex("02 01   7b")		# 123
		asn1_details = ASN1Tools.safe_decode(der_data, asn1_spec = RestrictiveASN1Model())
		self.assertEqual(asn1_details.flags, set())
		self.assertEqual(asn1_details.asn1, Integer(123))
		self.assertEqual(asn1_details.generic_asn1, None)
		self.assertEqual(asn1_details.encoded_der, None)
		self.assertEqual(asn1_details.original_der, der_data)
		self.assertEqual(asn1_details.model_index, 0)

	def test_safedecode_ok_model_fallback(self):
		der_data = bytes.fromhex("02 02 01c8")		# 456
		asn1_details = ASN1Tools.safe_decode(der_data, asn1_spec = (RestrictiveASN1Model(), BasicASN1Model()))
		self.assertEqual(asn1_details.flags, set([ "fallback" ]))
		self.assertEqual(asn1_details.tail, bytes())
		self.assertEqual(asn1_details.asn1, Integer(456))
		self.assertEqual(asn1_details.generic_asn1, None)
		self.assertEqual(asn1_details.encoded_der, None)
		self.assertEqual(asn1_details.original_der, der_data)
		self.assertEqual(asn1_details.model_index, 1)

	def test_safedecode_fail_model_generic(self):
		der_data = bytes.fromhex("02 02 01c8")		# 456
		asn1_details = ASN1Tools.safe_decode(der_data, asn1_spec = RestrictiveASN1Model())
		self.assertEqual(asn1_details.flags, set([ "unexpected_type" ]))
		self.assertEqual(asn1_details.tail, bytes())
		self.assertEqual(asn1_details.asn1, None)
		self.assertEqual(asn1_details.generic_asn1, Integer(456))
		self.assertEqual(asn1_details.encoded_der, None)
		self.assertEqual(asn1_details.original_der, der_data)
		self.assertEqual(asn1_details.model_index, None)

	def test_safedecode_ok_trailing_data(self):
		der_data = bytes.fromhex("02 02 01c8 aabb")		# 456
		asn1_details = ASN1Tools.safe_decode(der_data, asn1_spec = BasicASN1Model())
		self.assertEqual(asn1_details.flags, set([ "trailing_data" ]))
		self.assertEqual(asn1_details.asn1, Integer(456))
		self.assertEqual(asn1_details.tail, bytes.fromhex("aa bb"))
		self.assertEqual(asn1_details.generic_asn1, None)
		self.assertEqual(asn1_details.encoded_der, None)
		self.assertEqual(asn1_details.original_der, der_data)
		self.assertEqual(asn1_details.model_index, 0)

	def test_safedecode_ok_nonder(self):
		der_data = bytes.fromhex("02 02 007b")		# 123 non-DER
		asn1_details = ASN1Tools.safe_decode(der_data, asn1_spec = RestrictiveASN1Model())
		self.assertEqual(asn1_details.flags, set([ "non_der" ]))
		self.assertEqual(asn1_details.asn1, Integer(123))
		self.assertEqual(asn1_details.tail, bytes())
		self.assertEqual(asn1_details.generic_asn1, None)
		self.assertEqual(asn1_details.encoded_der, bytes.fromhex("02 01 7b"))
		self.assertEqual(asn1_details.original_der, der_data)
		self.assertEqual(asn1_details.model_index, 0)

	def test_safedecode_ok_nonder_trailing_data(self):
		der_data = bytes.fromhex("02 02 007b aabb")		# 123 non-DER
		asn1_details = ASN1Tools.safe_decode(der_data, asn1_spec = RestrictiveASN1Model())
		self.assertEqual(asn1_details.flags, set([ "non_der", "trailing_data" ]))
		self.assertEqual(asn1_details.asn1, Integer(123))
		self.assertEqual(asn1_details.tail, bytes.fromhex("aabb"))
		self.assertEqual(asn1_details.generic_asn1, None)
		self.assertEqual(asn1_details.encoded_der, bytes.fromhex("02 01 7b"))
		self.assertEqual(asn1_details.original_der, der_data)
		self.assertEqual(asn1_details.model_index, 0)

	def test_safedecode_wrong_type(self):
		der_data = bytes.fromhex("04 06 666f6f626172")		# "foobar" OctetString
		asn1_details = ASN1Tools.safe_decode(der_data, asn1_spec = RestrictiveASN1Model())
		self.assertEqual(asn1_details.flags, set([ "unexpected_type" ]))
		self.assertEqual(asn1_details.asn1, None)
		self.assertEqual(asn1_details.tail, bytes())
		self.assertEqual(asn1_details.generic_asn1, OctetString(b"foobar"))
		self.assertEqual(asn1_details.encoded_der, None)
		self.assertEqual(asn1_details.original_der, der_data)
		self.assertEqual(asn1_details.model_index, None)

	def test_safedecode_non_encodable(self):
		der_data = bytes.fromhex("17 11 31 36 30 32 31 37 31 30 32 37 30 30 2b 30 31 30 30")		# UTCTime("160217102700+0100")
		asn1_details = ASN1Tools.safe_decode(der_data)
		self.assertEqual(asn1_details.flags, set([ "non_encodable" ]))
		self.assertEqual(asn1_details.asn1, UTCTime("160217102700+0100"))
		self.assertEqual(asn1_details.tail, bytes())
		self.assertEqual(asn1_details.generic_asn1, None)
		self.assertEqual(asn1_details.encoded_der, None)
		self.assertEqual(asn1_details.original_der, der_data)
		self.assertEqual(asn1_details.model_index, 0)



#	x509sak - The X.509 Swiss Army Knife white-hat certificate toolkit
#	Copyright (C) 2018-2020 Johannes Bauer
#
#	This file is part of x509sak.
#
#	x509sak is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	x509sak is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with x509sak; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import datetime
import calendar
from pyasn1.type.useful import GeneralizedTime
from x509sak.Tools import DateTimeTools
from x509sak.estimate.BaseEstimator import BaseEstimator
from x509sak.estimate import JudgementCode, Verdict, Commonness, Compatibility
from x509sak.estimate.Judgement import SecurityJudgement, SecurityJudgements, RFCReference

@BaseEstimator.register
class CrtValiditySecurityEstimator(BaseEstimator):
	_ALG_NAME = "crt_validity"

	def _format_datetime(self, dt):
		return {
			"iso":		DateTimeTools.strfmt_iso(dt),
			"timet":	calendar.timegm(dt.utctimetuple()),
		}

	def analyze(self, certificate):
		not_before = certificate.valid_not_before
		not_after = certificate.valid_not_after
		is_ca = certificate.is_ca_certificate
		judgements = SecurityJudgements()

		if not_before is None:
			judgements += SecurityJudgement(JudgementCode.X509Cert_Body_Validity_NotBefore_Malformed, "'Not Before' timestamp is malformed. Certificate is always invalid.", bits = 0, compatibility = Compatibility.STANDARDS_DEVIATION)
			validity_days = 0
		elif not_after is None:
			judgements += SecurityJudgement(JudgementCode.X509Cert_Body_Validity_NotAfter_Malformed, "'Not After' timestamp is malformed. Certificate is always invalid.", bits = 0, compatibility = Compatibility.STANDARDS_DEVIATION)
			validity_days = 0
		else:
			now = datetime.datetime.utcnow()
			if not_before > not_after:
				judgements += SecurityJudgement(JudgementCode.X509Cert_Body_Validity_Status_NeverValid, "'Not Before' timestamp is greater than 'not after' timestamp. Certificate is always invalid.", bits = 0)
			elif now < not_before:
				judgements += SecurityJudgement(JudgementCode.X509Cert_Body_Validity_Status_NotYetValid, "Certificate is not yet valid, becomes valid in the future.", bits = 0, commonness = Commonness.UNUSUAL)
			elif now > not_after:
				judgements += SecurityJudgement(JudgementCode.X509Cert_Body_Validity_Status_Expired, "Certificate has expired.", bits = 0, commonness = Commonness.COMMON)
			else:
				judgements += SecurityJudgement(JudgementCode.X509Cert_Body_Validity_Status_CurrentlyValid, "Certificate is currently valid.", commonness = Commonness.COMMON)

			validity_days = ((not_after - not_before).total_seconds()) / 86400

			if not is_ca:
				margins = [ 2 * 365.25, 5 * 365.25, 7 * 365.25 ]
			else:
				margins = [ 12.5 * 365.25, 25 * 365.25, 30 * 365.25 ]

			# Add some error margin
			margins = [ margin + 14 for margin in margins ]

			crt_type = "CA" if is_ca else "non-CA"
			if validity_days < margins[0]:
				judgements += SecurityJudgement(JudgementCode.X509Cert_Body_Validity_Length_Conservative, "Lifetime is conservative for %s certificate." % (crt_type), commonness = Commonness.COMMON, verdict = Verdict.BEST_IN_CLASS)
			elif validity_days < margins[1]:
				judgements += SecurityJudgement(JudgementCode.X509Cert_Body_Validity_Length_Long, "Lifetime is long, but still acceptable for %s certificate." % (crt_type), commonness = Commonness.COMMON, verdict = Verdict.HIGH)
			elif validity_days < margins[2]:
				judgements += SecurityJudgement(JudgementCode.X509Cert_Body_Validity_Length_VeryLong, "Lifetime is very long for %s certificate." % (crt_type), commonness = Commonness.UNUSUAL, verdict = Verdict.MEDIUM)
			else:
				judgements += SecurityJudgement(JudgementCode.X509Cert_Body_Validity_Length_ExceptionallyLong, "Lifetime is exceptionally long for %s certificate." % (crt_type), commonness = Commonness.HIGHLY_UNUSUAL, verdict = Verdict.WEAK)

			if (not_before < datetime.datetime(2050, 1, 1, 0, 0, 0)) and isinstance(certificate.asn1["tbsCertificate"]["validity"]["notBefore"].getComponent(), GeneralizedTime):
				standard = RFCReference(rfcno = 5280, sect = "4.1.2.5", verb = "MUST", text = "CAs conforming to this profile MUST always encode certificate validity dates through the year 2049 as UTCTime; certificate validity dates in 2050 or later MUST be encoded as GeneralizedTime.")
				judgements += SecurityJudgement(JudgementCode.X509Cert_Body_Validity_NotBefore_InvalidType, "GeneralizedTime used for 'not before' validity timestamp although earlier than year 2050.", compatibility = Compatibility.STANDARDS_DEVIATION, standard = standard)
			if (not_after < datetime.datetime(2050, 1, 1, 0, 0, 0)) and isinstance(certificate.asn1["tbsCertificate"]["validity"]["notAfter"].getComponent(), GeneralizedTime):
				standard = RFCReference(rfcno = 5280, sect = "4.1.2.5", verb = "MUST", text = "CAs conforming to this profile MUST always encode certificate validity dates through the year 2049 as UTCTime; certificate validity dates in 2050 or later MUST be encoded as GeneralizedTime.")
				judgements += SecurityJudgement(JudgementCode.X509Cert_Body_Validity_NotAfter_InvalidType, "GeneralizedTime used for 'not after' validity timestamp although earlier than year 2050.", compatibility = Compatibility.STANDARDS_DEVIATION, standard = standard)

		return {
			"not_before":		self._format_datetime(not_before) if not_before is not None else None,
			"not_after":		self._format_datetime(not_after) if not_after is not None else None,
			"validity_days":	validity_days,
			"security":			judgements,
		}



#	x509sak - The X.509 Swiss Army Knife white-hat certificate toolkit
#	Copyright (C) 2018-2018 Johannes Bauer
#
#	This file is part of x509sak.
#
#	x509sak is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	x509sak is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with x509sak; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

from x509sak.tests import BaseTest
from x509sak.PassphraseGenerator import PassphraseGenerator

class PassphraseGeneratorTests(BaseTest):
	def test_passphrase_integers(self):
		for bits in range(1, 10):
			maxvalue = 2 ** bits
			for _ in range(100):
				intvalue = PassphraseGenerator.rand_int(bits)
				self.assertGreaterEqual(intvalue, 0)
				self.assertLess(intvalue, maxvalue)

	def test_max_radix_bits(self):
		for i in range(1, 20):
			self.assertEqual(PassphraseGenerator.max_digits_in_radix(i, 2), i)
			self.assertEqual(PassphraseGenerator.max_digits_in_radix(i, 4), (i + 1) // 2)
			self.assertEqual(PassphraseGenerator.max_digits_in_radix(i, 8), (i + 2) // 3)
			self.assertEqual(PassphraseGenerator.max_digits_in_radix(i, 256), (i + 7) // 8)

		self.assertEqual(PassphraseGenerator.max_digits_in_radix(10, 26), 3)
		self.assertEqual(PassphraseGenerator.max_digits_in_radix(20, 26), 5)
		self.assertEqual(PassphraseGenerator.max_digits_in_radix(30, 26), 7)
		self.assertEqual(PassphraseGenerator.max_digits_in_radix(40, 26), 9)
		self.assertEqual(PassphraseGenerator.max_digits_in_radix(50, 26), 11)
		self.assertEqual(PassphraseGenerator.max_digits_in_radix(60, 26), 13)
		self.assertEqual(PassphraseGenerator.max_digits_in_radix(70, 26), 15)
		self.assertEqual(PassphraseGenerator.max_digits_in_radix(140, 26), 30)
		self.assertEqual(PassphraseGenerator.max_digits_in_radix(150, 26), 32)
		self.assertEqual(PassphraseGenerator.max_digits_in_radix(230, 26), 49)
		self.assertEqual(PassphraseGenerator.max_digits_in_radix(240, 26), 52)
		self.assertEqual(PassphraseGenerator.max_digits_in_radix(500, 26), 107)

	def test_passphrase_gen(self):
		for ppg in [ PassphraseGenerator.all_ascii(), PassphraseGenerator.non_ambiguous(), PassphraseGenerator.pronouncible() ]:
			passphrase = ppg.gen_passphrase(20)
			for _ in range(100):
				new_passphrase = ppg.gen_passphrase(20)
				self.assertEqual(len(passphrase), len(new_passphrase))



#	x509sak - The X.509 Swiss Army Knife white-hat certificate toolkit
#	Copyright (C) 2018-2018 Johannes Bauer
#
#	This file is part of x509sak.
#
#	x509sak is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	x509sak is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with x509sak; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import sys
from x509sak import CertificatePool, X509Certificate
from x509sak.BaseAction import BaseAction
from x509sak.Exceptions import InvalidUsageException
from x509sak.PrivateKeyStorage import PrivateKeyStorage, PrivateKeyStorageForm
from x509sak.OpenSSLTools import OpenSSLTools
from x509sak.PassphraseGenerator import PassphraseGenerator

class ActionBuildChain(BaseAction):
	def __init__(self, cmdname, args):
		BaseAction.__init__(self, cmdname, args)
		if (self._args.private_key is not None) and (self._args.outform != "pkcs12"):
			raise InvalidUsageException("A private key will only be exported with a PKCS#12 output file.")
		if self._args.outform != "pkcs12":
			if self._args.pkcs12_legacy_crypto:
				raise InvalidUsageException("Specifying PKCS#12 legacy crypto while not using PKCS#12 output format has no effect.")
			if self._args.pkcs12_no_passphrase:
				raise InvalidUsageException("Specifying no PKCS#12 passphrase while not using PKCS#12 output format has no effect.")
			if self._args.pkcs12_passphrase_file is not None:
				raise InvalidUsageException("Specifying a PKCS#12 passphrase while not using PKCS#12 output format has no effect.")
		else:
			if (self._args.pkcs12_legacy_crypto or self._args.pkcs12_no_passphrase or self._args.pkcs12_passphrase_file is not None) and (self._args.private_key is None):
				raise InvalidUsageException("Specifying any PKCS#12 passphrase/crypto options when not including a private key makes no sense.")

		self._pool = CertificatePool()
		self._load_truststore()
		if self._args.inform == "pem":
			certs = X509Certificate.read_pemfile(self._args.crtfile)
			if not self._args.dont_trust_crtfile:
				self._pool.add_certificates(certs)
			cert = certs[0]
		elif self._args.inform == "der":
			cert = X509Certificate.read_derfile(self._args.crtfile)
		else:
			raise NotImplementedError("inform", self._args.inform)
		chain = self._pool.find_chain(cert)
		if (chain.root is None) and (not self._args.allow_partial_chain):
			raise InvalidUsageException("Could not build full chain for certificate %s and partial chain matches are disallowed." % (self._args.crtfile))

		if self._args.outform == "rootonly":
			if chain.root is None:
				print("Root certificate output requested, but none found.", file = sys.stderr)
				sys.exit(1)
			certs = [ chain.root ]
		elif self._args.outform == "intermediates":
			certs = list(chain.chain)
		elif self._args.outform in [ "fullchain", "multifile", "pkcs12" ]:
			certs = list(chain.full_chain)
		elif self._args.outform == "all-except-root":
			certs = list(chain.full_chain)
			if chain.root is not None:
				certs = certs[:-1]
		else:
			raise NotImplementedError("outform", self._args.outform)

		if not self._args.order_leaf_to_root:
			certs = certs[::-1]

		for (cid, cert) in enumerate(certs):
			self._log.debug("Cert %d: %s", cid, cert)

		if self._args.outform == "multifile":
			for (cid, cert) in enumerate(certs):
				filename = self._args.outfile % (cid)
				with open(filename, "w") as f:
					print(cert.to_pem_data(), file = f)
		elif self._args.outform == "pkcs12":
			if self._args.private_key is not None:
				if self._args.pkcs12_no_passphrase:
					passphrase = None
				else:
					if self._args.pkcs12_passphrase_file is not None:
						with open(self._args.pkcs12_passphrase_file) as f:
							passphrase = f.readline()
					else:
						passphrase = PassphraseGenerator.non_ambiguous().gen_passphrase(80)
						print("Passphrase: %s" % (passphrase), file = sys.stderr)
			else:
				passphrase = None
			private_key_storage = PrivateKeyStorage(PrivateKeyStorageForm.PEM_FILE, filename = self._args.private_key) if (self._args.private_key is not None) else None
			pkcs12 = OpenSSLTools.create_pkcs12(certificates = certs, private_key_storage = private_key_storage, modern_crypto = not self._args.pkcs12_legacy_crypto, passphrase = passphrase)
			if self._args.outfile is None:
				# Write to stdout
				sys.stdout.buffer.write(pkcs12)
			else:
				# Write to file, binary
				with open(self._args.outfile, "wb") as f:
					f.write(pkcs12)
		else:
			if self._args.outfile is not None:
				with open(self._args.outfile, "w") as f:
					self._print_certs(f, certs)
			else:
				self._print_certs(sys.stdout, certs)

	@staticmethod
	def _print_certs(f, certs):
		for cert in certs:
			print(cert.to_pem_data(), file = f)

	def _load_truststore(self):
		self._pool.load_sources(self._args.ca_source)
		self._log.debug("Loaded a total of %d unique certificates in trust store.", self._pool.certificate_count)



#	x509sak - The X.509 Swiss Army Knife white-hat certificate toolkit
#	Copyright (C) 2018-2020 Johannes Bauer
#
#	This file is part of x509sak.
#
#	x509sak is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	x509sak is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with x509sak; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import collections
import pyasn1.error
import pyasn1.type.char
import pyasn1.codec.der.encoder
import pyasn1.codec.der.decoder
from pyasn1.type.char import UTF8String
from x509sak.OID import OID, OIDDB
from x509sak.Exceptions import InvalidInputException
from x509sak.StringParser import StringParser

_RFC2253_StringParser = StringParser(escape_chars = "\\+\"<>;,=")

class RelativeDistinguishedName():
	_RDNItem = collections.namedtuple("RDNItem", [ "oid", "derdata", "asn1", "decodable", "printable", "printable_value" ])

	def __init__(self, rdn_list):
		assert(isinstance(rdn_list, (tuple, list)))
		assert(all(isinstance(value[0], OID) for value in rdn_list))
		assert(all(isinstance(value[1], bytes) for value in rdn_list))
		if len(rdn_list) == 0:
			raise InvalidInputException("Empty RDN is not permitted.")
		self._rdn_list = tuple(self._decode_item(oid, derdata) for (oid, derdata) in rdn_list)

	@classmethod
	def create(cls, *key_values):
		def encode(text):
			asn1 = UTF8String(text)
			derdata = pyasn1.codec.der.encoder.encode(asn1)
			return derdata
		if (len(key_values) % 2) != 0:
			raise InvalidInputException("Length of key/values must be evenly divisible by two.")
		rdn_list = [ ]
		for (key, value) in zip(key_values[ : : 2], key_values[1 : : 2]):
			assert(isinstance(key, str))
			assert(isinstance(value, (str, bytes)))
			try:
				oid = OIDDB.RDNTypes.inverse(key)
			except KeyError:
				raise InvalidInputException("Cannot create RDN with key '%s', cannot look up OID for it." % (key))
			rdn_list.append((oid, encode(value)))
		return cls(rdn_list)

	@classmethod
	def from_asn1(cls, asn1):
		rdn_elements = [ ]
		for attribute_type_and_value in asn1:
			attribute_type_oid = OID.from_asn1(attribute_type_and_value[0])
			attribute_value = bytes(attribute_type_and_value[1])
			rdn_elements.append((attribute_type_oid, attribute_value))
		return cls(rdn_elements)

	@classmethod
	def _decode_item(cls, oid, derdata):
		try:
			(decoded_attribute_value, tail) = pyasn1.codec.der.decoder.decode(derdata)
			if len(tail) != 0:
				# Refuse to decode if there's trailing data.
				decoded_attribute_value = None
		except pyasn1.error.PyAsn1Error:
			decoded_attribute_value = None

		printable = isinstance(decoded_attribute_value, (pyasn1.type.char.UTF8String, pyasn1.type.char.PrintableString, pyasn1.type.char.IA5String, pyasn1.type.char.TeletexString, pyasn1.type.char.BMPString, pyasn1.type.char.UniversalString))
		if printable:
			# Use the string representation
			printable_value = str(decoded_attribute_value)
		else:
			# Use the bytes representation
			printable_value = "#" + derdata.hex()

		return cls._RDNItem(oid = oid, derdata = derdata, asn1 = decoded_attribute_value, decodable = decoded_attribute_value is not None, printable = printable, printable_value = printable_value)

	@property
	def component_cnt(self):
		return len(self._rdn_list)

	@property
	def oidkey(self):
		return tuple(rdn_item.oid for rdn_item in self._rdn_list)

	def has_component(self, oid):
		assert(isinstance(oid, OID))
		return any(oid == item.oid for item in self._rdn_list)

	def get_value(self, oid):
		assert(isinstance(oid, OID))
		for item in self._rdn_list:
			if item.oid == oid:
				return item
		return None

	@classmethod
	def from_rfc2253_str(cls, string):
		keyvalues = [ ]
		for item in _RFC2253_StringParser.split(string, control_char = "+", reassemble = True):
			keyvalue = _RFC2253_StringParser.split(item, control_char = "=", reassemble = True)
			if len(keyvalue) != 2:
				raise InvalidInputException("RDN string item has %d members, expected 2." % (len(keyvalue)))
			(key, value) = keyvalue
			keyvalues.append(key)
			keyvalues.append(value)
		return cls.create(*keyvalues)

	@property
	def rfc2253_str(self):
		def char_needs_hex_escaping(char):
			try:
				char.encode("ascii")
				codepoint = ord(char)
				return not (32 <= codepoint < 127)
			except UnicodeEncodeError:
				return True

		def escape(text):
			if isinstance(text, str):
				escaped_text = [ ]
				if text.startswith("#"):
					escaped_text.append("\\#")
					text = text[1:]
				elif text.startswith(" "):
					escaped_text.append("\\ ")
					text = text[1:]
				spc_end = text.endswith(" ")
				if spc_end:
					text = text[:-1]
				escape_chars = set("\\+\"<>;,")
				for char in text:
					if char in escape_chars:
						escaped_text.append("\\%s" % (char))
					elif char_needs_hex_escaping(char):
						escaped_text += [ "\\%02X" % (c) for c in char.encode("utf-8") ]
					else:
						escaped_text.append(char)
				if spc_end:
					escaped_text.append("\\ ")
				return "".join(escaped_text)
			else:
				# bytes handling
				return "#" + text.hex()
		return "+".join("%s=%s" % (OIDDB.RDNTypes.get(item.oid, item.oid), escape(item.printable_value or "/")) for item in reversed(self._rdn_list))

	@property
	def pretty_str(self):
		def escape(text):
			if (" " in text) or ("\"" in text) or ("," in text):
				return "\"%s\"" % (text.replace("\"", "\\\""))
			else:
				return text
		return ", ".join("%s = %s" % (OIDDB.RDNTypes.get(item.oid, item.oid), escape(item.printable_value or "/")) for item in sorted(self._rdn_list))

	def dump(self, indent = ""):
		for rdn in self:
			print("%s%s" % (indent, rdn.pretty_str))

	def __iter__(self):
		return iter(self._rdn_list)

	def _cmpkey(self):
		return ((item.oid, item.printable_value if item.printable else item.asn1) for item in self._rdn_list)

	def __hash__(self):
		return hash(tuple(sorted(self._cmpkey())))

	def __eq__(self, other):
		return sorted(self._cmpkey()) == sorted(other._cmpkey())

	def __neq__(self, other):
		return not (self == other)

	def __str__(self):
		return "RDN<%s>" % (self.rfc2253_str)

class DistinguishedName():
	def __init__(self, rdns):
		assert(isinstance(rdns, (list, tuple)))
		assert(all(isinstance(rdn, RelativeDistinguishedName) for rdn in rdns))
		self._rdns = tuple(rdns)

	@property
	def rdn_count(self):
		return len(self._rdns)

	@property
	def empty(self):
		return self.rdn_count == 0

	def get_all(self, oid):
		assert(isinstance(oid, OID))
		return [ rdn for rdn in self._rdns if rdn.has_component(oid) ]

	@classmethod
	def from_asn1(cls, asn1):
		# asn1 is of type rfc2459.Name and contains the RDNSequence
		rdn_sequence = asn1[0]

		rdns = [ ]
		for rdn in rdn_sequence:
			rdns.append(RelativeDistinguishedName.from_asn1(rdn))

		dn = cls(rdns)
		return dn

	@classmethod
	def from_rfc2253_str(cls, string):
		rdns = [ ]
		for rdn_parsed in _RFC2253_StringParser.split(string, control_char = ",", reassemble = True):
			rdns.append(RelativeDistinguishedName.from_rfc2253_str(rdn_parsed))
		return cls(rdns)

	@property
	def rfc2253_str(self):
		return ",".join(rdn.rfc2253_str for rdn in reversed(self._rdns))

	@property
	def pretty_str(self):
		return ", ".join(rdn.pretty_str for rdn in self._rdns)

	def dump(self, indent = ""):
		for rdn in self:
			rdn.dump(indent = indent)

	def __iter__(self):
		return iter(self._rdns)

	def __eq__(self, other):
		return self._rdns == other._rdns

	def __neq__(self, other):
		return not (self == other)

	def __lt__(self, other):
		return self._rdns < other._rdns

	def __hash__(self):
		return hash(self._rdns)

	def __str__(self):
		return "DN<%s>" % (self.rfc2253_str)



#	x509sak - The X.509 Swiss Army Knife white-hat certificate toolkit
#	Copyright (C) 2018-2020 Johannes Bauer
#
#	This file is part of x509sak.
#
#	x509sak is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	x509sak is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with x509sak; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

from x509sak.tests import BaseTest
from x509sak.Tools import ValidationTools

class ValidationToolsTests(BaseTest):
	def test_emails_valid(self):
		self.assertTrue(ValidationTools.validate_email_address("foo@bar.de"))
		self.assertTrue(ValidationTools.validate_email_address("x@y"))
		self.assertTrue(ValidationTools.validate_email_address("x@d.y"))
		self.assertTrue(ValidationTools.validate_email_address("x@a.d.y"))
		self.assertTrue(ValidationTools.validate_email_address("0@a.d.y"))
		self.assertTrue(ValidationTools.validate_email_address("foo@bar.co.uk"))
		self.assertTrue(ValidationTools.validate_email_address("foo.bar@bar.co.uk"))
		self.assertTrue(ValidationTools.validate_email_address("foo_bar@bar.co.uk"))

	def test_emails_invalid(self):
		self.assertFalse(ValidationTools.validate_email_address(""))
		self.assertFalse(ValidationTools.validate_email_address("@"))
		self.assertFalse(ValidationTools.validate_email_address("a"))
		self.assertFalse(ValidationTools.validate_email_address("abc@"))
		self.assertFalse(ValidationTools.validate_email_address("@abc"))
		self.assertFalse(ValidationTools.validate_email_address("abc@.abc"))
		self.assertFalse(ValidationTools.validate_email_address("abc@abc."))
		self.assertFalse(ValidationTools.validate_email_address("abc@ abc"))
		self.assertFalse(ValidationTools.validate_email_address("abc@ abc"))

	def test_domains_valid(self):
		self.assertTrue(ValidationTools.validate_domainname("bar.de"))
		self.assertTrue(ValidationTools.validate_domainname("y"))
		self.assertTrue(ValidationTools.validate_domainname("d.y"))
		self.assertTrue(ValidationTools.validate_domainname("a.d.y"))
		self.assertTrue(ValidationTools.validate_domainname("a.d.y"))
		self.assertTrue(ValidationTools.validate_domainname("bar.co.uk"))

	def test_domains_invalid(self):
		self.assertFalse(ValidationTools.validate_domainname(""))
		self.assertFalse(ValidationTools.validate_domainname("."))
		self.assertFalse(ValidationTools.validate_domainname(".foo"))
		self.assertFalse(ValidationTools.validate_domainname("foo."))
		self.assertFalse(ValidationTools.validate_domainname("foo bar"))

	def test_uri_invalid(self):
		self.assertFalse(ValidationTools.validate_uri("/foo/bar"))
		self.assertFalse(ValidationTools.validate_uri("moo.com/foobar"))
		self.assertFalse(ValidationTools.validate_uri("foo.de:80/jfisoud"))
		self.assertFalse(ValidationTools.validate_uri("://foo"))
		self.assertFalse(ValidationTools.validate_uri("://foo/"))
		self.assertFalse(ValidationTools.validate_uri("http:../example.com"))

	def test_uri_valid(self):
		self.assertTrue(ValidationTools.validate_uri("http://a"))
		self.assertTrue(ValidationTools.validate_uri("http://a.a/"))
		self.assertTrue(ValidationTools.validate_uri("http://a/a"))
		self.assertTrue(ValidationTools.validate_uri("http://a/a/"))
		self.assertTrue(ValidationTools.validate_uri("http://a/a/?"))
		self.assertTrue(ValidationTools.validate_uri("http://a/a/?a"))
		self.assertTrue(ValidationTools.validate_uri("http://de.wikipedia.org/wiki/Uniform_Resource_Identifier"))
		self.assertTrue(ValidationTools.validate_uri("http://de.wikipedia.org/wiki/Uniform_Resource_Identifier/"))
		self.assertTrue(ValidationTools.validate_uri("ftp://ftp.is.co.za/rfc/rfc1808.txt"))
		self.assertTrue(ValidationTools.validate_uri("file:///C:/Users/Benutzer/Desktop/Uniform%20Resource%20Identifier.html"))
		self.assertTrue(ValidationTools.validate_uri("file:///etc/fstab"))
		self.assertTrue(ValidationTools.validate_uri("geo:48.33,14.122;u=22.5"))
		self.assertTrue(ValidationTools.validate_uri("ldap://[2001:db8::7]/c=GB?objectClass?one"))
		self.assertTrue(ValidationTools.validate_uri("gopher://gopher.floodgap.com"))
		self.assertTrue(ValidationTools.validate_uri("mailto:John.Doe@example.com"))
		self.assertTrue(ValidationTools.validate_uri("sip:911@pbx.mycompany.com"))
		self.assertTrue(ValidationTools.validate_uri("news:comp.infosystems.www.servers.unix"))
		self.assertTrue(ValidationTools.validate_uri("data:text/plain;charset=iso-8859-7,%be%fa%be"))
		self.assertTrue(ValidationTools.validate_uri("tel:+1-816-555-1212"))
		self.assertTrue(ValidationTools.validate_uri("telnet://192.0.2.16:80/"))
		self.assertTrue(ValidationTools.validate_uri("urn:oasis:names:specification:docbook:dtd:xml:4.1.2"))
		self.assertTrue(ValidationTools.validate_uri("git://github.com/rails/rails.git"))
		self.assertTrue(ValidationTools.validate_uri("crid://broadcaster.com/movies/BestActionMovieEver"))
		self.assertTrue(ValidationTools.validate_uri("http://aa.bbbb-c.dd/eeeeee"))

	def test_domainname_template_valid(self):
		self.assertEqual(ValidationTools.validate_domainname_template("bar.de")[0], ValidationTools.DomainnameTemplateValidationResult.Valid)
		self.assertEqual(ValidationTools.validate_domainname_template("y")[0], ValidationTools.DomainnameTemplateValidationResult.Valid)
		self.assertEqual(ValidationTools.validate_domainname_template("d.y")[0], ValidationTools.DomainnameTemplateValidationResult.Valid)
		self.assertEqual(ValidationTools.validate_domainname_template("a.d.y")[0], ValidationTools.DomainnameTemplateValidationResult.Valid)
		self.assertEqual(ValidationTools.validate_domainname_template("a.d.y")[0], ValidationTools.DomainnameTemplateValidationResult.Valid)
		self.assertEqual(ValidationTools.validate_domainname_template("bar.co.uk")[0], ValidationTools.DomainnameTemplateValidationResult.Valid)

		self.assertEqual(ValidationTools.validate_domainname_template("*.co.uk")[0], ValidationTools.DomainnameTemplateValidationResult.Valid)
		self.assertEqual(ValidationTools.validate_domainname_template("foo*.co.uk")[0], ValidationTools.DomainnameTemplateValidationResult.Valid)
		self.assertEqual(ValidationTools.validate_domainname_template("*foo.co.uk")[0], ValidationTools.DomainnameTemplateValidationResult.Valid)
		self.assertEqual(ValidationTools.validate_domainname_template("foo.foo*.co.uk")[0], ValidationTools.DomainnameTemplateValidationResult.Valid)
		self.assertEqual(ValidationTools.validate_domainname_template("foo.*foo.co.uk")[0], ValidationTools.DomainnameTemplateValidationResult.Valid)
		self.assertEqual(ValidationTools.validate_domainname_template("*.xn--foo.co.uk")[0], ValidationTools.DomainnameTemplateValidationResult.Valid)

	def test_domainname_template_invalid(self):
		self.assertEqual(ValidationTools.validate_domainname_template("*foo*.foo.co.uk")[0], ValidationTools.DomainnameTemplateValidationResult.Valid)
		self.assertEqual(ValidationTools.validate_domainname_template("**.foo.co.uk")[0], ValidationTools.DomainnameTemplateValidationResult.Valid)
		self.assertEqual(ValidationTools.validate_domainname_template("f$oo.foo.co.uk")[0], ValidationTools.DomainnameTemplateValidationResult.InvalidCharacter)
		self.assertEqual(ValidationTools.validate_domainname_template("f*$oo.foo.co.uk")[0], ValidationTools.DomainnameTemplateValidationResult.InvalidCharacter)
		self.assertEqual(ValidationTools.validate_domainname_template("foo.*.co.uk")[0], ValidationTools.DomainnameTemplateValidationResult.FullWildcardNotLeftmost)
		self.assertEqual(ValidationTools.validate_domainname_template("foo.co.uk.*")[0], ValidationTools.DomainnameTemplateValidationResult.FullWildcardNotLeftmost)
		self.assertEqual(ValidationTools.validate_domainname_template("foo*.co*.uk")[0], ValidationTools.DomainnameTemplateValidationResult.MoreThanOneWildcard)
		self.assertEqual(ValidationTools.validate_domainname_template("foo.xn--foo*.uk")[0], ValidationTools.DomainnameTemplateValidationResult.WildcardInInternationalDomain)

	def test_domainname_template_match(self):
		self.assertTrue(ValidationTools.validate_domainname_template_match("foo.com", "FOO.cOm"))
		self.assertTrue(ValidationTools.validate_domainname_template_match("foo.CoM", "FOO.cOm"))
		self.assertTrue(ValidationTools.validate_domainname_template_match("*.com", "FOO.cOm"))
		self.assertTrue(ValidationTools.validate_domainname_template_match("*.foo.com", "wWw.FOO.cOm"))
		self.assertTrue(ValidationTools.validate_domainname_template_match("*blubb.foo.com", "wWwblubb.FOO.cOm"))
		self.assertTrue(ValidationTools.validate_domainname_template_match("blubb*.foo.com", "bLubbmoo.FOO.cOm"))
		self.assertTrue(ValidationTools.validate_domainname_template_match("muh*blubb.foo.com", "mUhwWwblubb.FOO.cOm"))
		self.assertTrue(ValidationTools.validate_domainname_template_match("muh.f*o.com", "mUh.FOO.cOm"))
		self.assertTrue(ValidationTools.validate_domainname_template_match("muh.f*o.com", "mUh.FO.cOm"))
		self.assertTrue(ValidationTools.validate_domainname_template_match("muh.f*o.com", "mUh.FabcdefO.cOm"))
		self.assertTrue(ValidationTools.validate_domainname_template_match("muh.fo*.com", "mUh.FOobar.cOm"))

	def test_domainname_template_no_match(self):
		self.assertFalse(ValidationTools.validate_domainname_template_match("foo.com", "foo.de"))
		self.assertFalse(ValidationTools.validate_domainname_template_match("*.com", "www.FOO.com"))
		self.assertFalse(ValidationTools.validate_domainname_template_match("*.foo.com", "FOO.cOm"))
		self.assertFalse(ValidationTools.validate_domainname_template_match("*blubb.foo.com", "wWw.FOO.cOm"))
		self.assertFalse(ValidationTools.validate_domainname_template_match("blubb*.foo.com", "www.FOO.cOm"))
		self.assertFalse(ValidationTools.validate_domainname_template_match("muh*blubb.foo.com", "mUh.FOO.cOm"))
		self.assertFalse(ValidationTools.validate_domainname_template_match("muh*blubb.foo.com", "blubb.FOO.cOm"))
		self.assertFalse(ValidationTools.validate_domainname_template_match("muh*blubb.foo.com", "abcblubb.FOO.cOm"))
		self.assertFalse(ValidationTools.validate_domainname_template_match("muh.f*o.com", "mUh.web.cOm"))



#	x509sak - The X.509 Swiss Army Knife white-hat certificate toolkit
#	Copyright (C) 2018-2020 Johannes Bauer
#
#	This file is part of x509sak.
#
#	x509sak is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	x509sak is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with x509sak; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

from x509sak.AlgorithmDB import Cryptosystems
from x509sak.estimate.BaseEstimator import BaseEstimator
from x509sak.Exceptions import LazyDeveloperException
from x509sak.CurveDB import CurveNotFoundException
from x509sak.estimate.Judgement import SecurityJudgement, SecurityJudgements, JudgementCode, Compatibility, Commonness
from x509sak.PublicKey import RSAPublicKey
from x509sak.estimate.DERValidator import DERValidator

@BaseEstimator.register
class PublicKeyEstimator(BaseEstimator):
	_ALG_NAME = "pubkey"
	_RSA_PUBKEY_DER_VALIDATOR = DERValidator.create_inherited("X509Cert_PublicKey_RSA", validation_subject = "Certificate RSA Public Key")

	def _analyze_pubkey_encoding_rsa(self, pubkey):
		judgements = SecurityJudgements()
		asn1_details = pubkey.decoding_details
		judgements += self._RSA_PUBKEY_DER_VALIDATOR.validate(asn1_details)
		return judgements

	def _analyze_pubkey_encoding(self, pubkey):
		if isinstance(pubkey, RSAPublicKey):
			return self._analyze_pubkey_encoding_rsa(pubkey)

	def _error_curve_not_found(self, certificate, exception):
		judgements = SecurityJudgements()
		judgements += SecurityJudgement(JudgementCode.X509Cert_PublicKey_ECC_DomainParameters_Name_UnkownName, "Certificate public key relies on unknown elliptic curve: %s Conservatively estimating broken security." % (str(exception)), bits = 0, compatibility = Compatibility.LIMITED_SUPPORT, commonness = Commonness.HIGHLY_UNUSUAL)
		result = {
			"pubkey_alg":	None,
			"pretty":		"Unrecognized elliptic curve: %s" % (str(exception)),
			"security":		judgements,
		}
		return result

	def _error_public_key_type_unknown(self, certificate, decoding_error):
		judgements = SecurityJudgements()
		judgements += SecurityJudgement(JudgementCode.X509Cert_PublicKey_UnknownKeyType, "Certificate public key uses unknown key system: %s Conservatively estimating broken security." % (decoding_error), bits = 0, compatibility = Compatibility.LIMITED_SUPPORT, commonness = Commonness.HIGHLY_UNUSUAL)
		result = {
			"pubkey_alg":	None,
			"pretty":		"Undecodable public key: %s" % (decoding_error),
			"security":		judgements,
		}
		return result

	def analyze(self, certificate):
		try:
			pubkey = certificate.pubkey
			if not pubkey.key_decodable:
				return self._error_public_key_type_unknown(certificate, pubkey.key_decoding_error)
		except CurveNotFoundException as e:
			return self._error_curve_not_found(certificate, e)

		result = {
			"pubkey_alg":	pubkey.pk_alg.value.name,
			"security":		SecurityJudgements(),
		}

		if pubkey.key.malformed:
			result["pretty"] = "%s with malformed encoding" % (pubkey.pk_alg.name)
		else:
			if pubkey.pk_alg.value.cryptosystem == Cryptosystems.RSA:
				result["pretty"] = "RSA with %d bit modulus" % (pubkey.n.bit_length())
				result.update(self.algorithm("rsa").analyze(pubkey))
			elif pubkey.pk_alg.value.cryptosystem == Cryptosystems.DSA:
				result["pretty"] = "DSA with %d bit modulus and %d bit output" % (pubkey.p.bit_length(), pubkey.q.bit_length())
				result.update(self.algorithm("dsa").analyze(pubkey))
			elif pubkey.pk_alg.value.cryptosystem == Cryptosystems.ECC_ECDSA:
				if pubkey.curve_source == "namedCurve":
					result["pretty"] = "ECC on %s" % (pubkey.curve.name)
				elif pubkey.curve_source == "specifiedCurve":
					result["pretty"] = "ECC on specified %s" % (str(pubkey.curve))
				else:
					result["pretty"] = "ECC on %s" % (str(pubkey.curve_source))
				result.update(self.algorithm("ecc").analyze(pubkey))
			elif pubkey.pk_alg.value.cryptosystem == Cryptosystems.ECC_EdDSA:
				result["pretty"] = "EdDSA on %s" % (pubkey.curve.name)
				result.update(self.algorithm("ecc").analyze(pubkey))
			else:
				raise LazyDeveloperException(NotImplemented, pubkey.pk_alg.value.cryptosystem)

		result["security"] += self._analyze_pubkey_encoding(certificate.pubkey.key)
		return result



#	x509sak - The X.509 Swiss Army Knife white-hat certificate toolkit
#	Copyright (C) 2018-2018 Johannes Bauer
#
#	This file is part of x509sak.
#
#	x509sak is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	x509sak is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with x509sak; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import tempfile
from x509sak.WorkDir import WorkDir
from x509sak.tests import BaseTest, ResourceFileLoader
from x509sak.SubprocessExecutor import SubprocessExecutor
from x509sak.X509Certificate import X509Certificate

class CmdLineTestsBuildChain(BaseTest):
	def test_root_only_out_default(self):
		with ResourceFileLoader("certs/ok/johannes-bauer-root.pem") as certfile:
			output = self._run_x509sak([ "buildchain", certfile ]).stdout
		self.assertOcurrences(output, b"-----BEGIN CERTIFICATE-----", 1)
		crts = X509Certificate.from_pem_data(output)
		self.assertEqual(crts[0].subject.rfc2253_str, "CN=DST Root CA X3,O=Digital Signature Trust Co.")

	def test_root_notrust(self):
		with ResourceFileLoader("certs/ok/johannes-bauer-root.pem") as certfile:
			output = self._run_x509sak([ "buildchain", "--dont-trust-crtfile", certfile ]).stdout
		self.assertOcurrences(output, b"-----BEGIN CERTIFICATE-----", 1)
		crts = X509Certificate.from_pem_data(output)
		self.assertEqual(crts[0].subject.rfc2253_str, "CN=DST Root CA X3,O=Digital Signature Trust Co.")

	def test_interm_root_notrust(self):
		with ResourceFileLoader("certs/ok/johannes-bauer-intermediate.pem") as certfile:
			self._run_x509sak([ "buildchain", certfile ], success_return_codes = [ 1 ])
			output = self._run_x509sak([ "buildchain", "--allow-partial-chain", certfile ]).stdout
		self.assertOcurrences(output, b"-----BEGIN CERTIFICATE-----", 1)
		crts = X509Certificate.from_pem_data(output)
		self.assertEqual(crts[0].subject.rfc2253_str, "CN=Let's Encrypt Authority X3,O=Let's Encrypt,C=US")

		with ResourceFileLoader([ "certs/ok/johannes-bauer-root.pem" ], "certs/ok/johannes-bauer-intermediate.pem") as (searchdir, certfile):
			output = self._run_x509sak([ "buildchain", "-s", searchdir, certfile ]).stdout
		self.assertOcurrences(output, b"-----BEGIN CERTIFICATE-----", 2)
		crts = X509Certificate.from_pem_data(output.decode("ascii"))
		self.assertEqual(crts[0].subject.rfc2253_str, "CN=DST Root CA X3,O=Digital Signature Trust Co.")
		self.assertEqual(crts[1].subject.rfc2253_str, "CN=Let's Encrypt Authority X3,O=Let's Encrypt,C=US")

	def test_der_input(self):
		with tempfile.TemporaryDirectory() as tempdir, WorkDir(tempdir):
			crt = self._load_crt("ok/johannes-bauer-root")
			crt.write_derfile("root.der")
			output = self._run_x509sak([ "buildchain", "--inform", "der", "root.der" ]).stdout
			self.assertOcurrences(output, b"-----BEGIN CERTIFICATE-----", 1)
			crts = X509Certificate.from_pem_data(output.decode("ascii"))
			self.assertEqual(crts[0].subject.rfc2253_str, "CN=DST Root CA X3,O=Digital Signature Trust Co.")

	def test_root_only_out_rootonly(self):
		with ResourceFileLoader([ "certs/ok/johannes-bauer-root.pem", "certs/ok/johannes-bauer-intermediate.pem" ], "certs/ok/johannes-bauer.com.pem") as (searchdir, certfile):
			output = self._run_x509sak([ "buildchain", "-s", searchdir, "--outform", "rootonly", certfile ]).stdout
		self.assertOcurrences(output, b"-----BEGIN CERTIFICATE-----", 1)
		crts = X509Certificate.from_pem_data(output.decode("ascii"))
		self.assertEqual(crts[0].subject.rfc2253_str, "CN=DST Root CA X3,O=Digital Signature Trust Co.")

	def test_intermediate(self):
		with ResourceFileLoader([ "certs/ok/johannes-bauer-root.pem", "certs/ok/johannes-bauer-intermediate.pem" ], "certs/ok/johannes-bauer.com.pem") as (searchdir, certfile):
			output = self._run_x509sak([ "buildchain", "-s", searchdir, "--outform", "intermediates", certfile ]).stdout
		self.assertOcurrences(output, b"-----BEGIN CERTIFICATE-----", 1)
		crts = X509Certificate.from_pem_data(output.decode("ascii"))
		self.assertEqual(crts[0].subject.rfc2253_str, "CN=Let's Encrypt Authority X3,O=Let's Encrypt,C=US")

	def test_all_except_root(self):
		with ResourceFileLoader([ "certs/ok/johannes-bauer-root.pem", "certs/ok/johannes-bauer-intermediate.pem" ], "certs/ok/johannes-bauer.com.pem") as (searchdir, certfile):
			output = self._run_x509sak([ "buildchain", "-s", searchdir, "--outform", "all-except-root", certfile ]).stdout
		self.assertOcurrences(output, b"-----BEGIN CERTIFICATE-----", 2)
		crts = X509Certificate.from_pem_data(output.decode("ascii"))
		self.assertEqual(crts[0].subject.rfc2253_str, "CN=Let's Encrypt Authority X3,O=Let's Encrypt,C=US")
		self.assertEqual(crts[1].subject.rfc2253_str, "CN=johannes-bauer.com")

	def test_all_except_root_stdout(self):
		with tempfile.NamedTemporaryFile(prefix = "chain_", suffix = ".pem") as outfile, ResourceFileLoader([ "certs/ok/johannes-bauer-root.pem", "certs/ok/johannes-bauer-intermediate.pem" ], "certs/ok/johannes-bauer.com.pem") as (searchdir, certfile):
			output = self._run_x509sak([ "buildchain", "-s", searchdir, "--outform", "all-except-root", "--outfile", outfile.name, certfile ]).stdout
			self.assertOcurrences(output, b"-----BEGIN CERTIFICATE-----", 0)
			crts = X509Certificate.read_pemfile(outfile.name)
			self.assertEqual(crts[0].subject.rfc2253_str, "CN=Let's Encrypt Authority X3,O=Let's Encrypt,C=US")
			self.assertEqual(crts[1].subject.rfc2253_str, "CN=johannes-bauer.com")

	def test_multifile_all_except_root(self):
		with tempfile.TemporaryDirectory() as tempdir, WorkDir(tempdir), ResourceFileLoader([ "certs/ok/johannes-bauer-root.pem", "certs/ok/johannes-bauer-intermediate.pem" ], "certs/ok/johannes-bauer.com.pem") as (searchdir, certfile):
			self._run_x509sak([ "buildchain", "-s", searchdir, "--outform", "multifile", "--outfile", "outcrt%02d.pem", certfile ])
			self.assertEqual(X509Certificate.read_pemfile("outcrt00.pem")[0].subject.rfc2253_str, "CN=DST Root CA X3,O=Digital Signature Trust Co.")
			self.assertEqual(X509Certificate.read_pemfile("outcrt01.pem")[0].subject.rfc2253_str, "CN=Let's Encrypt Authority X3,O=Let's Encrypt,C=US")
			self.assertEqual(X509Certificate.read_pemfile("outcrt02.pem")[0].subject.rfc2253_str, "CN=johannes-bauer.com")

			self._run_x509sak([ "buildchain", "-s", searchdir, "--outform", "multifile", "--order-leaf-to-root", "--outfile", "rev_outcrt%02d.pem", certfile ])
			self.assertEqual(X509Certificate.read_pemfile("rev_outcrt00.pem")[0].subject.rfc2253_str, "CN=johannes-bauer.com")
			self.assertEqual(X509Certificate.read_pemfile("rev_outcrt01.pem")[0].subject.rfc2253_str, "CN=Let's Encrypt Authority X3,O=Let's Encrypt,C=US")
			self.assertEqual(X509Certificate.read_pemfile("rev_outcrt02.pem")[0].subject.rfc2253_str, "CN=DST Root CA X3,O=Digital Signature Trust Co.")

	def test_pkcs12(self):
		with tempfile.NamedTemporaryFile(suffix = ".p12") as p12file, ResourceFileLoader([ "certs/ok/johannes-bauer-root.pem", "certs/ok/johannes-bauer-intermediate.pem" ], "certs/ok/johannes-bauer.com.pem") as (searchdir, certfile):
			self._run_x509sak([ "buildchain", "-s", searchdir, "--outform", "pkcs12", "--outfile", p12file.name, certfile ])
			output = SubprocessExecutor([ "openssl", "pkcs12", "-in", p12file.name, "-passin", "pass:" ]).run().stdout
			self.assertOcurrences(output, b"-----BEGIN CERTIFICATE-----", 3)

	def test_cmd_errors(self):
		self._run_x509sak([ "buildchain", "--private-key", "foo.key", "x509sak/tests/data/johannes-bauer-root.crt" ], success_return_codes = [ 1 ])
		self._run_x509sak([ "buildchain", "--pkcs12-legacy-crypto", "x509sak/tests/data/johannes-bauer-root.crt" ], success_return_codes = [ 1 ])
		self._run_x509sak([ "buildchain", "--pkcs12-no-passphrase", "x509sak/tests/data/johannes-bauer-root.crt" ], success_return_codes = [ 1 ])
		self._run_x509sak([ "buildchain", "--pkcs12-passphrase-file", "pass.txt", "x509sak/tests/data/johannes-bauer-root.crt" ], success_return_codes = [ 1 ])
		self._run_x509sak([ "buildchain", "--outform", "pkcs12", "--pkcs12-passphrase-file", "pass.txt", "x509sak/tests/data/johannes-bauer-root.crt" ], success_return_codes = [ 1 ])
		self._run_x509sak([ "buildchain", "--outform", "pkcs12", "--pkcs12-legacy-crypto", "pass.txt", "x509sak/tests/data/johannes-bauer-root.crt" ], success_return_codes = [ 1 ])

	def test_cmd_no_root_found(self):
		self._run_x509sak([ "buildchain", "--allow-partial-chain", "--outform", "rootonly", "x509sak/tests/data/johannes-bauer-intermediate.crt" ], success_return_codes = [ 1 ])

	def test_pkcs12_stdout(self):
		with ResourceFileLoader([ "certs/ok/johannes-bauer-root.pem", "certs/ok/johannes-bauer-intermediate.pem" ], "certs/ok/johannes-bauer.com.pem") as (searchdir, certfile):
			pkcs12 =  self._run_x509sak([ "buildchain", "-s", searchdir, "--outform", "pkcs12", certfile ]).stdout
			output = SubprocessExecutor([ "openssl", "pkcs12", "-passin", "pass:" ], stdin = pkcs12).run().stdout
			self.assertOcurrences(output, b"-----BEGIN CERTIFICATE-----", 3)

	def test_pkcs12_passphrase_specify(self):
		with tempfile.NamedTemporaryFile(mode = "w", suffix = ".txt") as passfile, ResourceFileLoader("certs/ok/ecc_secp256r1.pem", "privkey/ok/ecc_secp256r1.pem") as (certfile, keyfile):
			print("foobar", file = passfile)
			passfile.flush()
			pkcs12 =  self._run_x509sak([ "buildchain", "--outform", "pkcs12", "--private-key", keyfile, "--pkcs12-passphrase-file", passfile.name, certfile ]).stdout

			# Fails with wrong passphrase
			SubprocessExecutor([ "openssl", "pkcs12", "-passin", "-nodes", "pass:" ], stdin = pkcs12, success_return_codes = [ 1 ]).run()
			SubprocessExecutor([ "openssl", "pkcs12", "-passin", "-nodes", "pass:abcdef" ], stdin = pkcs12, success_return_codes = [ 1 ]).run()

			# Works with right passphrase
			output = SubprocessExecutor([ "openssl", "pkcs12", "-nodes", "-passin", "pass:foobar" ], stdin = pkcs12).run().stdout
			self.assertOcurrences(output, b"-----BEGIN CERTIFICATE-----", 1)
			self.assertOcurrences(output, b"-----BEGIN PRIVATE KEY-----", 1)

	def test_pkcs12_passphrase_autogen(self):
		with ResourceFileLoader("certs/ok/ecc_secp256r1.pem", "privkey/ok/ecc_secp256r1.pem") as (certfile, keyfile):
			result = self._run_x509sak([ "buildchain", "--outform", "pkcs12", "--private-key", keyfile, certfile ])

			pkcs12 = result.stdout
			stderr = result.stderr_text.rstrip("\r\n")
			self.assertTrue(stderr.startswith("Passphrase: "))
			passphrase = stderr[12:]

			# Fails with wrong passphrase
			SubprocessExecutor([ "openssl", "pkcs12", "-passin", "-nodes", "pass:" ], stdin = pkcs12, success_return_codes = [ 1 ]).run()
			SubprocessExecutor([ "openssl", "pkcs12", "-passin", "-nodes", "pass:abcdef" ], stdin = pkcs12, success_return_codes = [ 1 ]).run()

			# Works with right passphrase
			output = SubprocessExecutor([ "openssl", "pkcs12", "-nodes", "-passin", "pass:" + passphrase ], stdin = pkcs12).run().stdout
			self.assertOcurrences(output, b"-----BEGIN CERTIFICATE-----", 1)
			self.assertOcurrences(output, b"-----BEGIN PRIVATE KEY-----", 1)

	def test_multi_intermediate(self):
		with ResourceFileLoader("certs/ok/multi_intermediate/root.crt", "certs/ok/multi_intermediate/interm1.crt", "certs/ok/multi_intermediate/interm2.crt", "certs/ok/multi_intermediate/client.crt") as (root_file, interm1_file, interm2_file, client_file):
			# Assert that without any intermediate certificate, building a chain fails
			self._run_x509sak([ "buildchain", "-s", root_file, client_file ], success_return_codes = [ 1 ])

			interm1_crt = X509Certificate.read_pemfile(interm1_file)[0]
			interm2_crt = X509Certificate.read_pemfile(interm2_file)[0]

			# With intermediate 1, it works.
			output = self._run_x509sak([ "buildchain", "-s", root_file, "-s", interm1_file, client_file ]).stdout
			certs = X509Certificate.from_pem_data(output)
			self.assertEqual(len(certs), 3)
			self.assertEqual(certs[1], interm1_crt)

			# Similarly, with intermedaite 2, it works.
			output = self._run_x509sak([ "buildchain", "-s", root_file, "-s", interm2_file, client_file ]).stdout
			certs = X509Certificate.from_pem_data(output)
			self.assertEqual(len(certs), 3)
			self.assertEqual(certs[1], interm2_crt)

			# But if both are accepted, intemediate 2 wins (it's newer)
			output = self._run_x509sak([ "buildchain", "-s", root_file, "-s", interm1_file, "-s", interm2_file, client_file ]).stdout
			certs = X509Certificate.from_pem_data(output)
			self.assertEqual(len(certs), 3)
			self.assertEqual(certs[1], interm2_crt)



#	x509sak - The X.509 Swiss Army Knife white-hat certificate toolkit
#	Copyright (C) 2018-2018 Johannes Bauer
#
#	This file is part of x509sak.
#
#	x509sak is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	x509sak is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with x509sak; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import hashlib
from x509sak.PRNG import HashedPRNG
from x509sak.tests.BaseTest import BaseTest

class PRNGTests(BaseTest):
	def test_prng(self):
		prng = HashedPRNG(b"foobar")
		first = prng.get(128)
		prng.reset()
		second = prng.get(128)
		self.assertEqual(first, second)

	def test_pattern(self):
		data = HashedPRNG(b"foobar").get(1024)
		self.assertEqual(hashlib.md5(data).hexdigest(), "acc988f3788299f2bd603c38902ae0a7")
		data = HashedPRNG(b"barfoo").get(1024)
		self.assertEqual(hashlib.md5(data).hexdigest(), "fe06bbd9bf3988af7b3f2d12b20bd583")

	def test_beginning(self):
		data = HashedPRNG(b"foobar").get(1024)
		for l in range(len(data) + 1):
			part = HashedPRNG(b"foobar").get(l)
			self.assertEqual(part, data[:l])

	def test_twoparts(self):
		data = HashedPRNG(b"foobar").get(256)
		for l in range(len(data) + 1):
			prng = HashedPRNG(b"foobar")
			part1 = prng.get(l)
			part2 = prng.get(len(data) - l)
			self.assertEqual(data, part1 + part2)



#	x509sak - The X.509 Swiss Army Knife white-hat certificate toolkit
#	Copyright (C) 2018-2018 Johannes Bauer
#
#	This file is part of x509sak.
#
#	x509sak is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	x509sak is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with x509sak; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import tempfile
from x509sak.tests import BaseTest
from x509sak.WorkDir import WorkDir
from x509sak.SubprocessExecutor import SubprocessExecutor

class CmdLineTestsSignCSR(BaseTest):
	def test_sign_pubkey(self):
		with tempfile.TemporaryDirectory() as tempdir, WorkDir(tempdir):
			# Creata a CA
			self._run_x509sak([ "createca", "myCA" ])

			# Create a default CSR
			self._run_x509sak([ "createcsr", "client.key", "client.csr" ])

			# Sign the CSR
			self._run_x509sak([ "signcsr", "myCA", "client.csr", "client.crt" ])

			# Extract public key from private key
			pubkey_in = SubprocessExecutor([ "openssl", "ec", "-pubout", "-in", "client.key" ]).run().stdout

			# Extract public key from certificate
			pubkey_out = SubprocessExecutor([ "openssl", "x509", "-noout", "-pubkey", "-in", "client.crt" ]).run().stdout

			self.assertEqual(pubkey_in, pubkey_out)

	def test_take_override_subject(self):
		with tempfile.TemporaryDirectory() as tempdir, WorkDir(tempdir):
			# Creata a CA
			self._run_x509sak([ "createca", "myCA" ])

			# Create a CSR with a specific CN/OU
			self._run_x509sak([ "createcsr", "-s", "/CN=CSR_CN_Subject/OU=CSR_OU_Subject", "client.key", "client.csr" ])

			# Sign the CSR and take original subject
			self._run_x509sak([ "signcsr", "myCA", "client.csr", "client_original.crt" ])

			# Sign the CSR and override subject
			self._run_x509sak([ "signcsr", "-s", "/CN=Foobar", "myCA", "client.csr", "client_override.crt" ])

			# Get text representation of CSR and CRT
			csr_text = SubprocessExecutor([ "openssl", "req", "-noout", "-text", "-in", "client.csr" ]).run().stdout
			crt_orig_text = SubprocessExecutor([ "openssl", "x509", "-noout", "-text", "-in", "client_original.crt" ]).run().stdout
			crt_override_text = SubprocessExecutor([ "openssl", "x509", "-noout", "-text", "-in", "client_override.crt" ]).run().stdout

			# The subject is in the CSR and the CRT without any specification,
			# but not in the CRT where it was overriden
			self.assertIn(b"CSR_CN_Subject", csr_text)
			self.assertIn(b"CSR_OU_Subject", csr_text)
			self.assertIn(b"CSR_CN_Subject", crt_orig_text)
			self.assertIn(b"CSR_OU_Subject", crt_orig_text)
			self.assertNotIn(b"CSR_CN_Subject", crt_override_text)
			self.assertNotIn(b"CSR_OU_Subject", crt_override_text)

	def test_san_included(self):
		with tempfile.TemporaryDirectory() as tempdir, WorkDir(tempdir):
			# Creata a CA
			self._run_x509sak([ "createca", "myCA" ])

			# Create a CSR
			self._run_x509sak([ "createcsr", "client.key", "client.csr" ])

			# Sign the CSR with a SAN
			self._run_x509sak([ "signcsr", "--san-dns", "foobar.com", "--san-dns", "barfoo.com", "--san-ip", "11.22.33.44", "myCA", "client.csr", "client.crt" ])

			# Check that SAN ended up in the CRT
			crt_text = SubprocessExecutor([ "openssl", "x509", "-noout", "-text", "-in", "client.crt" ]).run().stdout
			self.assertIn(b"X509v3 extensions", crt_text)
			self.assertIn(b"foobar.com", crt_text)
			self.assertIn(b"barfoo.com", crt_text)

	def test_override_extensions(self):
		with tempfile.TemporaryDirectory() as tempdir, WorkDir(tempdir):
			# Creata a CA
			self._run_x509sak([ "createca", "myCA" ])

			# Create a CSR and pack it full of extensions
			self._run_x509sak([ "createcsr", "-t", "tls-client", "--san-dns", "fooname.com", "--san-ip", "11.22.33.44", "client.key", "client.csr" ])

			# Sign the CSR but don't put any extensions in the CRT
			self._run_x509sak([ "signcsr", "myCA", "client.csr", "client_none.crt" ])

			# Sign the CSR a second time and override extensions
			self._run_x509sak([ "signcsr", "-s", "/CN=Server", "-t", "tls-server", "myCA", "client.csr", "client_server.crt" ])

			# Sign the CSR a third time and override extensions and add a SAN
			self._run_x509sak([ "signcsr", "-s", "/CN=Server with SAN DNS", "-t", "tls-server", "--san-dns", "newdns.com", "myCA", "client.csr", "client_server_dns.crt" ])

			# Get text representation of CSR and CRT
			csr_text = SubprocessExecutor([ "openssl", "req", "-noout", "-text", "-in", "client.csr" ]).run().stdout
			crt_none_text = SubprocessExecutor([ "openssl", "x509", "-noout", "-text", "-in", "client_none.crt" ]).run().stdout
			crt_server_text = SubprocessExecutor([ "openssl", "x509", "-noout", "-text", "-in", "client_server.crt" ]).run().stdout
			crt_server_dns_text = SubprocessExecutor([ "openssl", "x509", "-noout", "-text", "-in", "client_server_dns.crt" ]).run().stdout

			# CSR contains a bunch of extensions
			self.assertIn(b"Requested Extensions", csr_text)
			self.assertIn(b"fooname.com", csr_text)
			self.assertIn(b"11.22.33.44", csr_text)
			self.assertIn(b"Netscape Cert Type", csr_text)
			self.assertIn(b"Basic Constraints", csr_text)
			self.assertIn(b"SSL Client", csr_text)

			# But the CRT does not contain any
			self.assertNotIn(b"X509v3 extensions", crt_none_text)
			self.assertNotIn(b"fooname.com", crt_none_text)
			self.assertNotIn(b"11.22.33.44", crt_none_text)
			self.assertNotIn(b"Netscape Cert Type", crt_none_text)
			self.assertNotIn(b"Basic Constraints", crt_none_text)

			# The cert with template contains some, though.
			self.assertIn(b"X509v3 extensions", crt_server_text)
			self.assertNotIn(b"fooname.com", crt_server_text)
			self.assertNotIn(b"11.22.33.44", crt_server_text)
			self.assertIn(b"Netscape Cert Type", crt_server_text)
			self.assertIn(b"Basic Constraints", crt_server_text)
			self.assertIn(b"SSL Server", crt_server_text)

			# And one more, the last one also contains the specified SAN
			self.assertIn(b"X509v3 extensions", crt_server_dns_text)
			self.assertNotIn(b"fooname.com", crt_server_dns_text)
			self.assertNotIn(b"11.22.33.44", crt_server_dns_text)
			self.assertIn(b"Netscape Cert Type", crt_server_dns_text)
			self.assertIn(b"Basic Constraints", crt_server_dns_text)
			self.assertIn(b"SSL Server", crt_server_dns_text)
			self.assertIn(b"newdns.com", crt_server_dns_text)



#	x509sak - The X.509 Swiss Army Knife white-hat certificate toolkit
#	Copyright (C) 2020-2020 Johannes Bauer
#
#	This file is part of x509sak.
#
#	x509sak is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	x509sak is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with x509sak; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

from x509sak.tests import BaseTest
from x509sak.StringParser import StringParser, StringParseException

class StringParserTests(BaseTest):
	def test_unescaped(self):
		self.assertEqual(list(StringParser(escape_chars = ";").parse("foo")), [ (0, "f"), (0, "o"), (0, "o") ])

	def test_escaped(self):
		self.assertEqual(list(StringParser(escape_chars = ";").parse(r"foo\;bar")), [ (0, "f"), (0, "o"), (0, "o"), (0, ";"), (0, "b"), (0, "a"), (0, "r")])

	def test_control(self):
		self.assertEqual(list(StringParser(escape_chars = ";").parse(r"foo;bar")), [ (0, "f"), (0, "o"), (0, "o"), (1, ";"), (0, "b"), (0, "a"), (0, "r")])

	def test_end_of_string(self):
		with self.assertRaises(StringParseException):
			list(StringParser(escape_chars = ";").parse("foo;bar\\"))

	def test_no_escape_expected(self):
		with self.assertRaises(StringParseException):
			list(StringParser(escape_chars = ";").parse(r"foo\bar"))

	def test_split_basic(self):
		sstr = StringParser(escape_chars = ";").split("foo;bar", control_char = ";")
		self.assertEqual(len(sstr), 2)
		self.assertEqual(sstr[0], [ (0, "f"), (0, "o"), (0, "o") ])
		self.assertEqual(sstr[1], [ (0, "b"), (0, "a"), (0, "r") ])

	def test_split_meta(self):
		sstr = StringParser(escape_chars = ";+").split("f+o;bar", control_char = ";")
		self.assertEqual(len(sstr), 2)
		self.assertEqual(sstr[0], [ (0, "f"), (1, "+"), (0, "o") ])
		self.assertEqual(sstr[1], [ (0, "b"), (0, "a"), (0, "r") ])

	def test_split_escaped_meta(self):
		sstr = StringParser(escape_chars = ";+").split("f\\+o;bar", control_char = ";")
		self.assertEqual(len(sstr), 2)
		self.assertEqual(sstr[0], [ (0, "f"), (0, "+"), (0, "o") ])
		self.assertEqual(sstr[1], [ (0, "b"), (0, "a"), (0, "r") ])

	def test_split_recombine_meta(self):
		sp = StringParser(escape_chars = ";+")
		sstr = sp.split("f+o;bar", control_char = ";")
		self.assertEqual(len(sstr), 2)
		self.assertEqual(sp.escape(sstr[0]), "f+o")
		self.assertEqual(sp.escape(sstr[1]), "bar")

	def test_split_recombine_escaped_meta(self):
		sp = StringParser(escape_chars = ";+")
		sstr = sp.split("f\\+o;bar", control_char = ";")
		self.assertEqual(len(sstr), 2)
		self.assertEqual(sp.escape(sstr[0]), "f\\+o")
		self.assertEqual(sp.escape(sstr[1]), "bar")

	def test_split_recombine_meta_reassemble(self):
		sp = StringParser(escape_chars = ";+")
		sstr = sp.split("f+o;bar", control_char = ";", reassemble = True)
		self.assertEqual(len(sstr), 2)
		self.assertEqual(sstr[0], "f+o")
		self.assertEqual(sstr[1], "bar")

	def test_split_recombine_escaped_meta_reassemble(self):
		sp = StringParser(escape_chars = ";+")
		sstr = sp.split("f\\+o;bar", control_char = ";", reassemble = True)
		self.assertEqual(len(sstr), 2)
		self.assertEqual(sstr[0], "f\\+o")
		self.assertEqual(sstr[1], "bar")



#	x509sak - The X.509 Swiss Army Knife white-hat certificate toolkit
#	Copyright (C) 2018-2018 Johannes Bauer
#
#	This file is part of x509sak.
#
#	x509sak is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	x509sak is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with x509sak; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

from x509sak.estimate.BaseEstimator import BaseEstimator
from x509sak.estimate import Verdict, Commonness
from x509sak.estimate.Judgement import SecurityJudgement

@BaseEstimator.register
class BitsSecurityEstimator(BaseEstimator):
	_ALG_NAME = "bits"

	def analyze(self, code, bits):
		if bits < 64:
			judgement = SecurityJudgement(code, "Breakable with little effort (commercial-off-the-shelf hardware).", bits = bits, verdict = Verdict.NO_SECURITY, commonness = Commonness.HIGHLY_UNUSUAL, prefix_topic = True)
		elif bits < 80:
			judgement = SecurityJudgement(code, "Probably breakable with specialized hardware (limited purpose computers).", bits = bits, verdict = Verdict.WEAK, commonness = Commonness.UNUSUAL, prefix_topic = True)
		elif bits < 104:
			judgement = SecurityJudgement(code, "Nontrivial to break, but comparatively weak.", bits = bits, verdict = Verdict.WEAK, commonness = Commonness.UNUSUAL, prefix_topic = True)
		elif bits < 160:
			# 128 Bit security level
			judgement = SecurityJudgement(code, "High level of security.", bits = bits, verdict = Verdict.HIGH, commonness = Commonness.COMMON, prefix_topic = True)
		elif bits < 224:
			# 192 Bit security level
			judgement = SecurityJudgement(code, "Very high level of security.", bits = bits, verdict = Verdict.HIGH, commonness = Commonness.COMMON, prefix_topic = True)
		else:
			# 256 bit security level
			judgement = SecurityJudgement(code, "Exceptionally high level of security.", bits = bits, verdict = Verdict.BEST_IN_CLASS, commonness = Commonness.COMMON, prefix_topic = True)
		return judgement



#	x509sak - The X.509 Swiss Army Knife white-hat certificate toolkit
#	Copyright (C) 2019-2020 Johannes Bauer
#
#	This file is part of x509sak.
#
#	x509sak is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	x509sak is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with x509sak; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import re
import collections
from .Enums import CipherSuite

def _dict_by_identifier(obj_list):
	return { obj.identifier: obj for obj in obj_list }

class VerboseCipherSuite():
	_PSEUDO_SUITES = set([
		CipherSuite.TLS_EMPTY_RENEGOTIATION_INFO_SCSV,
		CipherSuite.TLS_FALLBACK_SCSV,
	])

	class _CSElements():
		KeyExchange = collections.namedtuple("KeyExchange", [ "identifier", "algorithm", "pfs", "export" ])
		SigAlgorithm = collections.namedtuple("SigAlgorithm", [ "identifier", "algorithm" ])
		Cipher = collections.namedtuple("Cipher", [ "identifier", "cipher", "keylen", "opmode", "aead" ])
		PRF = collections.namedtuple("PRF", [ "identifier", "hashlen" ])

	_CIPHER_SUITE_COMPONENTS = {
		"key_exchange": _dict_by_identifier([
			_CSElements.KeyExchange(identifier = "RSA", algorithm = "RSA", pfs = False, export = False),
			_CSElements.KeyExchange(identifier = "RSA_EXPORT", algorithm = "RSA", pfs = False, export = True),
			_CSElements.KeyExchange(identifier = "DH", algorithm = "DH", pfs = False, export = False),
			_CSElements.KeyExchange(identifier = "DH_anon", algorithm = "DH", pfs = False, export = False),
			_CSElements.KeyExchange(identifier = "DH_anon_EXPORT", algorithm = "DH", pfs = False, export = True),
			_CSElements.KeyExchange(identifier = "DHE", algorithm = "DH", pfs = True, export = False),
			_CSElements.KeyExchange(identifier = "ECDH", algorithm = "ECDH", pfs = False, export = False),
			_CSElements.KeyExchange(identifier = "ECDH_anon", algorithm = "ECDH", pfs = False, export = False),
			_CSElements.KeyExchange(identifier = "ECDHE", algorithm = "ECDH", pfs = True, export = False),
			_CSElements.KeyExchange(identifier = "PSK", algorithm = "PSK", pfs = False, export = False),
			_CSElements.KeyExchange(identifier = "KRB5", algorithm = "KRB5", pfs = False, export = False),
			_CSElements.KeyExchange(identifier = "PGP_DHE", algorithm = "DH", pfs = True, export = False),
			_CSElements.KeyExchange(identifier = "PGP_RSA", algorithm = "RSA", pfs = False, export = False),
			_CSElements.KeyExchange(identifier = "KRB5_EXPORT", algorithm = "KRB5", pfs = False, export = True),	# PFS??
			_CSElements.KeyExchange(identifier = "NULL", algorithm = None, pfs = False, export = True),
		]),
		"sig_algorithm": _dict_by_identifier([
			_CSElements.SigAlgorithm(identifier = "RSA", algorithm = "RSA"),
			_CSElements.SigAlgorithm(identifier = "RSA_EXPORT", algorithm = "RSA"),
			_CSElements.SigAlgorithm(identifier = "DSS", algorithm = "DSS"),
			_CSElements.SigAlgorithm(identifier = "DSS_EXPORT", algorithm = "DSS"),
			_CSElements.SigAlgorithm(identifier = "PSK", algorithm = "PSK"),
			_CSElements.SigAlgorithm(identifier = "ECDSA", algorithm = "ECDSA"),
		]),
		"cipher": _dict_by_identifier([
			_CSElements.Cipher(identifier = "3DES_EDE_CBC", cipher = "DES", keylen = 168, opmode = "CBC", aead = False),
			_CSElements.Cipher(identifier = "AES_128_CBC", cipher = "AES", keylen = 128, opmode = "CBC", aead = False),
			_CSElements.Cipher(identifier = "AES_128_GCM", cipher = "AES", keylen = 128, opmode = "GCM", aead = True),
			_CSElements.Cipher(identifier = "AES_256_CBC", cipher = "AES", keylen = 256, opmode = "CBC", aead = False),
			_CSElements.Cipher(identifier = "AES_256_GCM", cipher = "AES", keylen = 256, opmode = "GCM", aead = True),
			_CSElements.Cipher(identifier = "CAMELLIA_128_CBC", cipher = "Camellia", keylen = 128, opmode = "CBC", aead = False),
			_CSElements.Cipher(identifier = "CAMELLIA_256_CBC", cipher = "Camellia", keylen = 256, opmode = "CBC", aead = False),
			_CSElements.Cipher(identifier = "CAST_CBC", cipher = "CAST", keylen = 128, opmode = "CBC", aead = False),
			_CSElements.Cipher(identifier = "DES40_CBC", cipher = "DES", keylen = 40, opmode = "CBC", aead = False),
			_CSElements.Cipher(identifier = "DES_CBC_40", cipher = "DES", keylen = 40, opmode = "CBC", aead = False),
			_CSElements.Cipher(identifier = "DES_CBC", cipher = "DES", keylen = 56, opmode = "CBC", aead = False),
			_CSElements.Cipher(identifier = "IDEA_CBC", cipher = "IDEA", keylen = 128, opmode = "CBC", aead = False),
			_CSElements.Cipher(identifier = "NULL", cipher = None, keylen = 0, opmode = None, aead = False),
			_CSElements.Cipher(identifier = "RC2_CBC_40", cipher = "RC2", keylen = 40, opmode = "CBC", aead = False),
			_CSElements.Cipher(identifier = "RC4_128", cipher = "RC4", keylen = 128, opmode = "STRM", aead = False),
			_CSElements.Cipher(identifier = "RC4_40", cipher = "RC4", keylen = 40, opmode = "STRM", aead = False),
			_CSElements.Cipher(identifier = "SEED_CBC", cipher = "SEED", keylen = 128, opmode = "CBC", aead = False),
			_CSElements.Cipher(identifier = "CHACHA20_POLY1305", cipher = "ChaCha20", keylen = 256, opmode = "STRM", aead = True),
		]),
		"prf": _dict_by_identifier([
			_CSElements.PRF(identifier = "RMD", hashlen = 160),
			_CSElements.PRF(identifier = "MD5", hashlen = 128),
			_CSElements.PRF(identifier = "SHA", hashlen = 160),
			_CSElements.PRF(identifier = "SHA256", hashlen = 256),
			_CSElements.PRF(identifier = "SHA384", hashlen = 384),
		]),
	}

	_CACHE = { }

	def __init__(self, ciphersuite_id):
		assert(isinstance(ciphersuite_id, CipherSuite))
		self._csid = ciphersuite_id
		self._pseudo_suite = self._csid in self._PSEUDO_SUITES

		self._key_exchange = None
		self._sig_algorithm = None
		self._cipher = None
		self._prf = None

		self._pfs = False

		if not self._pseudo_suite:
			regex = self._cipher_suite_regex
			result = regex.match(str(self._csid))
			if result is not None:
				result = result.groupdict()
				self._key_exchange = self._CIPHER_SUITE_COMPONENTS["key_exchange"].get(result["key_exchange"])
				self._sig_algorithm = self._CIPHER_SUITE_COMPONENTS["sig_algorithm"].get(result["sig_algorithm"])
				self._cipher = self._CIPHER_SUITE_COMPONENTS["cipher"].get(result["cipher"])
				self._prf = self._CIPHER_SUITE_COMPONENTS["prf"].get(result["prf"])

	@property
	def csid(self):
		return self._csid

	@property
	def key_exchange(self):
		"""Returns the key exchange algorithm."""
		return self._key_exchange

	@property
	def sig_algorithm(self):
		"""Returns the signature algorithm."""
		return self._sig_algorithm

	@property
	def cipher(self):
		"""Returns the cipher algorithm which is used for bulk encrytion."""
		return self._cipher

	@property
	def prf(self):
		"""Returns the pseudo random function."""
		return self._prf

	@property
	def provides_pfs(self):
		"""Boolean that indicates whether or not the cipher suite provides
		perfect forward secrecy (PFS)."""
		return (self.key_exchange is not None) and self.key_exchange.pfs

	@property
	def is_pseudo_suite(self):
		"""Boolean that indicates whether or not the cipher suite is a
		non-negotiable pseudo cipher suite."""
		return self._pseudo_suite

	def _cipher_suite_regex_part(self, partname, optional = False):
		elements = self._CIPHER_SUITE_COMPONENTS[partname].values()
		part = "_(?P<%s>" % (partname)
		part += "|".join(element.identifier for element in elements)
		part += ")"
		if optional:
			part = "(%s)?" % (part)
		return part

	@property
	def _cipher_suite_regex(self):
		regex = r"^CipherSuite\.TLS"
		regex += self._cipher_suite_regex_part("key_exchange")
		regex += self._cipher_suite_regex_part("sig_algorithm", optional = True)
		regex += "_WITH"
		regex += self._cipher_suite_regex_part("cipher")
		regex += self._cipher_suite_regex_part("prf")
		regex += "$"
		return re.compile(regex)

	def negotiate(self):
		if self._pseudo_suite:
			raise Exception("Will not handle pseudo cipher suite '%s'." % (str(self._pseudo_suite)))

	@classmethod
	def getsuite(cls, ciphersuite_id):
		"""Caches cipher suites so they do not have to be parsed every time.
		Otherwise identical to the regular constructor."""
		if ciphersuite_id not in cls._CACHE:
			cls._CACHE[ciphersuite_id] = VerboseCipherSuite(ciphersuite_id)
		return cls._CACHE[ciphersuite_id]

	def __str__(self):
		return "%s<Pseudo=%s, KEx = %s, Sig = %s, Cipher = %s, PRF = %s>" % (self.csid, self._pseudo_suite, self.key_exchange, self.sig_algorithm, self.cipher, self.prf)


class CipherSuiteDirectory():
	def __init__(self, csids = None):
		if csids is not None:
			self._csuites = { csid: VerboseCipherSuite.getsuite(csid) for csid in csids }
		else:
			self._csuites = { }
			for csid in CipherSuite:
				csuite = VerboseCipherSuite.getsuite(csid)
				if not csuite.is_pseudo_suite:
					self._csuites[csid] = csuite

	def filter(self, filterfnc):
		return CipherSuiteDirectory(csuite.csid for csuite in self if filterfnc(csuite))

	def filter_cipher(self, filterfnc):
		return self.filter(lambda csuite: (csuite.cipher is not None) and filterfnc(csuite.cipher))

	def filter_kex(self, filterfnc):
		return self.filter(lambda csuite: (csuite.key_exchange is not None) and filterfnc(csuite.key_exchange))

	def filter_prf(self, filterfnc):
		return self.filter(lambda csuite: (csuite.prf is not None) and filterfnc(csuite.prf))

	def filter_sig_algorithm(self, filterfnc):
		return self.filter(lambda csuite: (csuite.sig_algorithm is not None) and filterfnc(csuite.sig_algorithm))

	def kwfilter(self, cipher_name = None, cipher_keylen = None, cipher_opmode = None, kex_alg = None, kex_pfs = None, sig_alg = None):
		directory = self
		if cipher_name is not None:
			directory = directory.filter_cipher(lambda cipher: cipher.cipher == cipher_name)
		if cipher_keylen is not None:
			directory = directory.filter_cipher(lambda cipher: cipher.keylen == cipher_keylen)
		if cipher_opmode is not None:
			directory = directory.filter_cipher(lambda cipher: cipher.opmode == cipher_opmode)
		if sig_alg is not None:
			directory = directory.filter_sig_algorithm(lambda sig: sig.algorithm == sig_alg)
		if kex_alg is not None:
			directory = directory.filter_kex(lambda kex: kex.algorithm == kex_alg)
		if kex_pfs is not None:
			directory = directory.filter_kex(lambda kex: kex.pfs == kex_pfs)
		return directory

	def filter_secure(self):
		secure_ciphers = self
		secure_ciphers = secure_ciphers.filter_cipher(lambda cipher: cipher.keylen >= 128)
		secure_ciphers = secure_ciphers.filter_kex(lambda kex: kex.pfs and (not kex.export))
		secure_ciphers = secure_ciphers.filter_prf(lambda prf: prf.hashlen > 160)
		secure_ciphers = secure_ciphers.filter_sig_algorithm(lambda sig_algorithm: sig_algorithm.identifier in [ "ECDSA", "RSA" ])
		return secure_ciphers

	def __iter__(self):
		return iter(self._csuites.values())

	def __len__(self):
		return len(self._csuites)

	def dump(self):
		for (csid, suite) in self._csuites.items():
			print(suite)

	def __str__(self):
		return "CipherSuiteDirectory<%d>" % (len(self))



#	x509sak - The X.509 Swiss Army Knife white-hat certificate toolkit
#	Copyright (C) 2020-2020 Johannes Bauer
#
#	This file is part of x509sak.
#
#	x509sak is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	x509sak is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with x509sak; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import collections

class FlagChecker():
	_CheckResult = collections.namedtuple("CheckResult", [ "check_type", "reference", "reference_count", "flags" ])

	def __init__(self):
		self._complex_checks = [ ]
		self._may_not_have = set()
		self._must_have = set()
		self._may_have = set()

	def complex_check(self, flags, min_count = None, max_count = None):
		assert((min_count is not None) or (max_count is not None))
		assert((min_count is None) or (min_count >= 0))
		assert((max_count is None) or (max_count <= len(flags)))
		self._complex_checks.append((set(flags), min_count, max_count))
		return self

	def may_have(self, *flags):
		self._may_have |= set(flags)
		return self

	def may_not_have(self, *flags):
		self._may_not_have |= set(flags)
		return self

	def must_have(self, *flags):
		self._must_have |= set(flags)
		return self

	def check(self, flags):
		flag_set = set(flags)
		missing_flags = self._must_have - flag_set
		if len(missing_flags) > 0:
			yield self._CheckResult(check_type = "missing", reference = self._must_have, reference_count = None, flags = missing_flags)

		excess_flags = self._may_not_have & flag_set
		if len(excess_flags) > 0:
			yield self._CheckResult(check_type = "excess", reference = self._may_not_have, reference_count = None, flags = excess_flags)

		unusual_flags = flag_set - self._must_have - self._may_have - self._may_not_have
		for (reference, min_count, max_count) in self._complex_checks:
			unusual_flags = unusual_flags - reference
		if len(unusual_flags) > 0:
			yield self._CheckResult(check_type = "unusual", reference = None, reference_count = None, flags = unusual_flags)

		for (reference, min_count, max_count) in self._complex_checks:
			intersection = reference & flag_set
			count = len(intersection)
			if (min_count is not None) and (count < min_count):
				yield self._CheckResult(check_type = "complex_too_few", reference = reference, reference_count = min_count, flags = intersection)
			if (max_count is not None) and (count > min_count):
				yield self._CheckResult(check_type = "complex_too_many", reference = reference, reference_count = max_count, flags = intersection)
